from typing import TYPE_CHECKING, Optional

import pytest

from message_ix_models.model.structure import get_codelist
from message_ix_models.tools.wb import (
    assign_income_groups,
    get_income_group_codelist,
    make_map,
)

if TYPE_CHECKING:
    import sdmx.model.common


@pytest.fixture(scope="module")
def cl_ig() -> "sdmx.model.common.Codelist":
    return get_income_group_codelist()


def test_get_income_group_codelist(cl_ig: "sdmx.model.common.Codelist") -> None:
    def n(id) -> int:
        return len(cl_ig[id].child)

    # Groups counts are as expected and have the expected relationships
    assert 25 == n("LIC")
    assert 108 == n("MIC") == n("LMC") + n("UMC")
    assert 77 == n("HIC")
    # NB +1 is for VEN, which is not categorized
    assert n("WLD") == n("LIC") + n("MIC") + n("HIC") + 1

    # Annotations exist and have expected values
    assert (
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).HIC"
        == str(cl_ig["ABW"].get_annotation(id="wb-income-group").text)
    )
    assert "IDA" == str(cl_ig["AFG"].get_annotation(id="wb-lending-category").text)


REPLACE = (
    {  # index 0
        "HIC": "HIC",
        "UMC": "LMIC",
        "LMC": "LMIC",
        "LIC": "LMIC",
    },
)

EXP = {
    ("R12", "count", None): {
        "R12_AFR": "LIC",
        "R12_RCPA": "LMC",
        "R12_CHN": "HIC",
        "R12_EEU": "HIC",
        "R12_FSU": "UMC",
        "R12_LAM": "UMC",
        "R12_MEA": "HIC",
        "R12_NAM": "HIC",
        "R12_PAO": "HIC",
        "R12_PAS": "LMC",
        "R12_SAS": "LMC",
        "R12_WEU": "HIC",
    },
    ("R12", "population", None): {
        "R12_AFR": "LMC",
        "R12_RCPA": "LMC",
        "R12_CHN": "UMC",
        "R12_EEU": "HIC",
        "R12_FSU": "UMC",
        "R12_LAM": "UMC",
        "R12_MEA": "LMC",
        "R12_NAM": "HIC",
        "R12_PAO": "HIC",
        "R12_PAS": "UMC",
        "R12_SAS": "LMC",
        "R12_WEU": "HIC",
    },
    ("R12", "population", 0): {
        "R12_AFR": "LMIC",
        "R12_RCPA": "LMIC",
        "R12_CHN": "LMIC",
        "R12_EEU": "HIC",
        "R12_FSU": "LMIC",
        "R12_LAM": "LMIC",
        "R12_MEA": "LMIC",
        "R12_NAM": "HIC",
        "R12_PAO": "HIC",
        "R12_PAS": "LMIC",
        "R12_SAS": "LMIC",
        "R12_WEU": "HIC",
    },
}


@pytest.mark.parametrize(
    "nodes, method, replace",
    (
        ("R12", "population", None),
        ("R12", "count", None),
        ("R12", "population", 0),
    ),
)
def test_assign_income_groups(
    cl_ig: "sdmx.model.common.Codelist", nodes: str, method: str, replace: Optional[int]
) -> None:
    cl_node = get_codelist(f"node/{nodes}")

    # Function runs without error
    replace_arg = make_map(REPLACE[replace]) if replace is not None else None

    assign_income_groups(cl_node, cl_ig, method, replace=replace_arg)

    # Each node belongs to the expected income group
    for node, ig in EXP[nodes, method, replace].items():
        c_node = cl_node[node]
        assert str(c_node.get_annotation(id="wb-income-group").text).endswith(ig), (
            c_node,
            c_node.annotations,
        )


def test_make_map() -> None:
    # Function runs
    result = make_map(REPLACE[0])

    # Expected result
    assert {
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).HIC": "HIC",
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).LMC": "LMIC",
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).UMC": "LMIC",
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).LIC": "LMIC",
    } == result



"""Tests of :mod:`.tools`."""

import pandas as pd
import pytest
from genno import Computer

from message_ix_models.testing import GHA
from message_ix_models.tools.exo_data import prepare_computer
from message_ix_models.tools.iea.web import DIMS, generate_code_lists, load_data


class TestIEA_EWEB:
    @pytest.mark.parametrize("source", ("IEA_EWEB",))
    @pytest.mark.parametrize(
        "source_kw",
        (
            dict(
                provider="OECD", edition="2021", product=["CHARCOAL"], flow=["RESIDENT"]
            ),
            # All flows related to transport
            dict(
                provider="OECD",
                edition="2022",
                flow=[
                    "DOMESAIR",
                    "DOMESNAV",
                    "PIPELINE",
                    "RAIL",
                    "ROAD",
                    "TOTTRANS",
                    "TRNONSPE",
                    "WORLDAV",
                    "WORLDMAR",
                ],
            ),
            pytest.param(
                dict(provider="IEA", edition="2023", extra_kw="FOO"),
                marks=pytest.mark.xfail(raises=ValueError),
            ),
        ),
    )
    def test_prepare_computer(self, test_context, source, source_kw):
        # FIXME The following should be redundant, but appears mutable on GHA linux and
        #       Windows runners.
        test_context.model.regions = "R14"

        c = Computer()

        keys = prepare_computer(test_context, c, source, source_kw)

        # Preparation of data runs successfully
        result = c.get(keys[0])
        # print(result.to_string())

        # Data has the expected dimensions
        assert {"n", "y", "product", "flow"} == set(result.dims)

        # Data contain expected coordinates
        # NB cannot test total counts here because the fuzzed test data does not
        #    necessarily include ≥1 data point from each COUNTRY and TIME
        assert {"R14_AFR", "R14_WEU"} < set(result.coords["n"].data)
        assert {1980, 2018} < set(result.coords["y"].data)


# NB once there is a fuzzed version of the (IEA, 2023) data available, usage of this
#    variable can be replaced with list(FILES.keys())
PROVIDER_EDITION = (
    pytest.param(
        "IEA",
        "2023",
        marks=pytest.mark.xfail(GHA, reason="No fuzzed version of this data"),
    ),
    ("OECD", "2023"),
    ("OECD", "2022"),
    ("OECD", "2023"),
)


@pytest.mark.parametrize("provider, edition", PROVIDER_EDITION)
def test_load_data(test_context, tmp_path, provider, edition):
    # # Store in the temporary directory for this test
    # test_context.cache_path = tmp_path.joinpath("cache")

    result = load_data(provider, edition)

    # Result has the correct type
    assert isinstance(result, pd.DataFrame)

    # print(result.head().to_string())  # DEBUG
    # print(result.head(1).transpose())  # DEBUG

    # Data have the expected dimensions
    assert (set(DIMS) & {"Value"}) < set(result.columns)


@pytest.mark.parametrize("provider, edition", PROVIDER_EDITION)
def test_generate_code_lists(tmp_path, provider, edition):
    # generate_code_lists() runs
    generate_code_lists(provider, edition, tmp_path)


@pytest.mark.parametrize(
    "urn, N",
    (
        ("IEA:COUNTRY_IEA(2023)", 191),
        ("IEA:COUNTRY_OECD(2021)", 190),
        ("IEA:COUNTRY_OECD(2022)", 185),
        ("IEA:COUNTRY_OECD(2023)", 191),
        ("IEA:FLOW_IEA(2023)", 108),
        ("IEA:FLOW_OECD(2021)", 108),
        ("IEA:FLOW_OECD(2022)", 108),
        ("IEA:FLOW_OECD(2023)", 108),
        ("IEA:PRODUCT_IEA(2023)", 68),
        ("IEA:PRODUCT_OECD(2021)", 68),
        ("IEA:PRODUCT_OECD(2022)", 68),
        ("IEA:PRODUCT_OECD(2023)", 68),
    ),
)
def test_load_codelists(urn, N):
    from message_ix_models.util.sdmx import read

    # Code list can be read using its URN
    cl = read(urn)

    # Code list has the expected number of codes
    assert N == len(cl)



from shutil import copyfile

import pytest

from message_ix_models.tools.advance import (
    DIMS,
    LOCATION,
    advance_data,
    get_advance_data,
)
from message_ix_models.util import package_data_path


@pytest.fixture(scope="module")
def advance_test_data(session_context):
    # Copy test data from the package directory into the local data directory for
    # `test_context`. get_advance_data() only uses this file if :mod:`message_data` is
    # NOT installed.
    target = session_context.get_local_path(*LOCATION)
    target.parent.mkdir(parents=True)
    copyfile(package_data_path("test", *LOCATION), target)


pytestmark = pytest.mark.usefixtures("advance_test_data")


def test_get_advance_data(session_context):
    """Test :func:`.get_advance_data`."""
    # Returns a pd.Series with the expected index levels
    result = get_advance_data()
    assert DIMS == result.index.names

    # Returns a genno.Quantity with the expected units
    result = advance_data("Transport|Service demand|Road|Freight")
    assert {"[length]": 1, "[mass]": 1, "[time]": -1} == result.units.dimensionality



""":class:`ScenarioInfo` class."""
import logging
import re
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from itertools import product
from typing import TYPE_CHECKING, Dict, List, Optional

import pandas as pd
import pint
import sdmx.model.v21 as sdmx_model

from .ixmp import parse_url
from .sdmx import eval_anno

if TYPE_CHECKING:
    from message_ix import Scenario

log = logging.getLogger(__name__)


# TODO: use kw_only=True once python 3.10 is oldest supported version
@dataclass()
class ScenarioInfo:
    """Information about a |Scenario| object.

    Code that prepares data for a target Scenario can accept a ScenarioInfo instance.
    This avoids the need to create or load an actual Scenario, which can be slow under
    some conditions.

    ScenarioInfo objects can also be used (for instance, by :func:`.apply_spec`) to
    describe the contents of a Scenario *before* it is created.

    ScenarioInfo objects have the following convenience attributes:

    .. autosummary::
       ~ScenarioInfo.set
       io_units
       is_message_macro
       N
       units_for
       Y
       y0
       yv_ya

    Parameters
    ----------
    scenario_obj : message_ix.Scenario
        If given, :attr:`.set` is initialized from this existing scenario.

    Examples
    --------
    Iterating over an instance gives "model", "scenario", "version" and the values of
    the respective attributes:
    >>> si = ScenarioInfo.from_url("model name/scenario name#123")
    >>> dict(si)
    {'model': 'model name', 'scenario': 'scenario name', 'version': 123}

    See also
    --------
    .Spec
    """

    # TODO: give this field kw_only=False once python 3.10 is the minimum version
    # Parameters for initialization only
    scenario_obj: InitVar[Optional["Scenario"]] = field(default=None)
    empty: InitVar[bool] = False

    platform_name: Optional[str] = None

    #: Model name; equivalent to :attr:`.TimeSeries.model`.
    model: Optional[str] = None

    #: Scenario name; equivalent to :attr:`.TimeSeries.scenario`.
    scenario: Optional[str] = None

    #: Scenario version; equivalent to :attr:`.TimeSeries.version`.
    version: Optional[int] = None

    #: Elements of :mod:`ixmp`/:mod:`message_ix` sets.
    set: Dict[str, List] = field(default_factory=lambda: defaultdict(list))

    #: Elements of :mod:`ixmp`/:mod:`message_ix` parameters.
    par: Dict[str, pd.DataFrame] = field(default_factory=dict)

    #: First model year, if set, else ``Y[0]``.
    y0: int = -1

    #: :obj:`True` if a MESSAGE-MACRO scenario.
    is_message_macro: bool = False

    _yv_ya: Optional[pd.DataFrame] = None

    def __post_init__(self, scenario_obj: Optional["Scenario"], empty: bool):
        if not scenario_obj:
            return

        self.model = scenario_obj.model
        self.scenario = scenario_obj.scenario
        self.version = (
            None if scenario_obj.version is None else int(scenario_obj.version)
        )

        if empty:
            return

        # Copy structure (set contents)
        for name in scenario_obj.set_list():
            value = scenario_obj.set(name)
            try:
                self.set[name] = value.tolist()
            except AttributeError:
                self.set[name] = value  # pd.DataFrame for ≥2-D set; don't convert

        # Copy data for a limited set of parameters
        for name in ("duration_period",):
            self.par[name] = scenario_obj.par(name)

        self.is_message_macro = "PRICE_COMMODITY" in scenario_obj.par_list()

        # Computed once
        fmy = scenario_obj.cat("year", "firstmodelyear")
        self.y0 = int(fmy[0]) if len(fmy) else self.set["year"][0]

        self._yv_ya = scenario_obj.vintage_and_active_years()

    @classmethod
    def from_url(cls, url: str) -> "ScenarioInfo":
        """Create an instance using only an :attr:`url`."""
        result = cls()
        result.url = url
        return result

    @property
    def yv_ya(self):
        """:class:`pandas.DataFrame` with valid ``year_vtg``, ``year_act`` pairs."""
        if self._yv_ya is None:
            # - Cartesian product of all yv and ya.
            # - Convert to data frame.
            # - Filter only valid years.
            self._yv_ya = (
                pd.DataFrame(
                    product(self.set["year"], self.set["year"]),
                    columns=["year_vtg", "year_act"],
                )
                .query("@self.y0 <= year_vtg <= year_act")
                .reset_index(drop=True)
            )

        return self._yv_ya

    @property
    def N(self):
        """Elements of the set 'node'.

        See also
        --------
        .nodes_ex_world
        """
        return list(map(str, self.set["node"]))

    @property
    def Y(self):
        """Elements of the set 'year' that are >= the first model year."""
        return list(filter(lambda y: y >= self.y0, self.set["year"]))

    @property
    def url(self) -> str:
        """Identical to :attr:`.TimeSeries.url`."""
        return f"{self.model}/{self.scenario}#{self.version}"

    @url.setter
    def url(self, value):
        p, s = parse_url(value)
        self.platform_name = p.get("name")
        for k in "model", "scenario", "version":
            setattr(self, k, s.get(k))

    _path_re = [
        (re.compile(r"[/<>:\"\\\|\?\*]+"), "_"),
        (re.compile("#"), "_v"),
        (re.compile("__+"), "_"),
    ]

    @property
    def path(self) -> str:
        """A valid file system path name similar to :attr:`url`.

        Characters invalid in Windows paths are replaced with "_".
        """
        from functools import reduce

        return reduce(lambda s, e: e[0].sub(e[1], s), self._path_re, self.url)

    def update(self, other: "ScenarioInfo"):
        """Update with the set elements of `other`."""
        for name, data_list in other.set.items():
            self.set[name].extend(
                filter(lambda id: id not in self.set[name], data_list)
            )

        for name, data_frame in other.par.items():
            raise NotImplementedError("Merging parameter data")

    def __iter__(self):
        for k in "model", "scenario", "version":
            yield (k, getattr(self, k))

    def __repr__(self):
        return (
            f"<ScenarioInfo: {sum(len(v) for v in self.set.values())} code(s) in "
            f"{len(self.set)} set(s)>"
        )

    def units_for(self, set_name: str, id: str) -> pint.Unit:
        """Return the units associated with code `id` in MESSAGE set `set_name`.

        :mod:`ixmp` (or the sole :class:`~ixmp.backend.base.JDBCBackend`, as of v3.5.0)
        does not handle unit information for variables and equations (unlike parameter
        values), such as MESSAGE decision variables ``ACT``, ``CAP``, etc. In
        :mod:`message_ix_models` and :mod:`message_data`, the following conventions are
        (generally) followed:

        - The units of ``ACT`` and others are consistent for each ``technology``.
        - The units of ``COMMODITY_BALANCE``, ``STOCK``, ``commodity_stock``, etc. are
          consistent for each ``commodity``.

        Thus, codes for elements of these sets (e.g. :ref:`commodity-yaml`) can be used
        to carry the standard units for the corresponding quantities. :func:`units_for`
        retrieves these units, for use in model-building and reporting.

        .. todo:: Expand this discussion and transfer to the :mod:`message_ix` docs.

        See also
        --------
        io_units
        """
        try:
            idx = self.set[set_name].index(id)
        except ValueError:
            print(self.set[set_name])
            raise

        return eval_anno(self.set[set_name][idx], "units")

    def io_units(
        self, technology: str, commodity: str, level: Optional[str] = None
    ) -> pint.Unit:
        """Return units for the MESSAGE ``input`` or ``output`` parameter.

        These are implicitly determined as the ratio of:

        - The units for the origin (for ``input``) or destination `commodity`, per
          :meth:`.units_for`.
        - The units of activity for the `technology`.

        Parameters
        ----------
        level : str
            Placeholder for future functionality, i.e. to use different units per
            (commodity, level). Currently ignored. If given, a debug message is logged.

        Raises
        ------
        ValueError
            if either `technology` or `commodity` lack defined units.
        """
        if level is not None:
            log.debug(f"{level = } ignored")
        c = self.units_for("commodity", commodity)
        t = self.units_for("technology", technology)
        if None in (c, t):
            raise ValueError(
                "Cannot compute input/output units for: "
                f"commodity={commodity!r} [{c}] / technology={technology!r} [{t}]"
            )
        return c / t

    def year_from_codes(self, codes: List[sdmx_model.Code]):
        """Update using a list of `codes`.

        The following are updated:

        - :attr:`.set` ``year``
        - :attr:`.set` ``cat_year``, with the first model year.
        - :attr:`.par` ``duration_period``

        Any existing values are discarded.

        After this, the attributes :attr:`.y0` and :attr:`.Y` give the first model year
        and model years, respectively.

        Examples
        --------
        Get a particular code list, create a ScenarioInfo instance, and update using
        the codes:

        >>> years = get_codes("year/A")
        >>> info = ScenarioInfo()
        >>> info.year_from_codes(years)

        Use populated values:

        >>> info.y0
        2020
        >>> info.Y[:3]
        [2020, 2030, 2040]
        >>> info.Y[-3:]
        [2090, 2100, 2110]

        """
        from message_ix_models.util import eval_anno

        # Clear existing values
        if len(self.set["year"]):
            log.debug(f"Discard existing 'year' elements: {repr(self.set['year'])}")
            self.set["year"] = []
        if len(self.set["cat_year"]):
            log.debug(
                f"Discard existing 'cat_year' elements: {repr(self.set['cat_year'])}"
            )
            self.set["cat_year"] = []
        if "duration_period" in self.par:
            log.debug("Discard existing 'duration_period' elements")

        fmy_set = False
        duration_period: List[Dict] = []

        # TODO use sorted() here once supported by sdmx
        for code in codes:
            year = int(code.id)
            # Store the year
            self.set["year"].append(year)

            # Check for an annotation 'firstmodelyear: true'
            if eval_anno(code, "firstmodelyear"):
                if fmy_set:
                    # No coverage: data that triggers this should not be committed
                    raise ValueError(  # pragma: no cover
                        "≥2 periods are annotated firstmodelyear: true"
                    )

                self.y0 = year
                self.set["cat_year"].append(("firstmodelyear", year))
                fmy_set = True

            # Store the duration_period: either from an annotation, or computed vs. the
            # prior period
            duration_period.append(
                dict(
                    year=year,
                    value=eval_anno(code, "duration_period")
                    or (year - duration_period[-1]["year"]),
                    unit="y",
                )
            )

        # Store
        self.par["duration_period"] = pd.DataFrame(duration_period)


@dataclass
class Spec:
    """A specification for the structure of a model or variant.

    A Spec collects 3 :class:`.ScenarioInfo` instances at the attributes :attr:`.add`,
    :attr:`.remove`, and :attr:`.require`. This is the type that is accepted by
    :func:`.apply_spec`; :doc:`model-build` describes how a Spec is used to modify a
    |Scenario|. A Spec may also be used to express information about the target
    structure of data to be prepared; like ScenarioInfo, this can happen before the
    target Scenario exists.

    Spec also provides:

    - Dictionary-style access, e.g. ``s["add"]`` is equivalent to ``s.add.``.
    - Error checking; setting keys other than add/remove/require results in an error.
    - :meth:`.merge`, a helper method.
    """

    #: Structure to be added to a base scenario.
    add: ScenarioInfo = field(default_factory=ScenarioInfo)
    #: Structure to be removed from a base scenario.
    remove: ScenarioInfo = field(default_factory=ScenarioInfo)
    #: Structure that must be present in a base scenario.
    require: ScenarioInfo = field(default_factory=ScenarioInfo)

    # Dict-like features

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value: ScenarioInfo):
        if not hasattr(self, key):
            raise KeyError(key)
        setattr(self, key, value)

    def values(self):
        yield self.add
        yield self.remove
        yield self.require

    # Static methods

    @staticmethod
    def merge(a: "Spec", b: "Spec") -> "Spec":
        """Merge Specs `a` and `b` together.

        Returns a new Spec where each member is a union of the respective members of
        `a` and `b`.
        """
        result = Spec()

        for key in {"add", "remove", "require"}:
            result[key].update(a[key])
            result[key].update(b[key])

        return result



import logging
from copy import copy

from message_ix_models.tools.exo_data import (
    ExoDataSource,
    iamc_like_data_for_query,
    register_source,
)
from message_ix_models.util import (
    HAS_MESSAGE_DATA,
    package_data_path,
    private_data_path,
)

__all__ = [
    "SSPOriginal",
    "SSPUpdate",
]

log = logging.getLogger(__name__)


@register_source
class SSPOriginal(ExoDataSource):
    """Provider of exogenous data from the original SSP database.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    arguments:

    - `source`: Any value from :data:`.SSP_2017` or equivalent string, for instance
      "ICONICS:SSP(2017).2". The specific SSP for which data is returned is determined
      from the value.
    - `source_kw` including:

      - "model": one of:

          - IIASA GDP
          - IIASA-WiC POP
          - NCAR
          - OECD Env-Growth
          - PIK GDP-32

      - "measure": The measures available differ according to the model; see the source
        data for details.

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="ICONICS:SSP(2015).3",
    ...     source_kw=dict(measure="POP", model="IIASA-WiC POP"),
    ... )
    >>> result = computer.get(keys[0])
    """

    id = "SSP"

    #: One-to-one correspondence between "model" codes and date fragments in scenario
    #: codes.
    model_date = {
        "IIASA GDP": "130219",
        "IIASA-WiC POP": "130115",
        "NCAR": "130115",
        "OECD Env-Growth": "130325",
        "PIK GDP-32": "130424",
    }

    #: Replacements to apply when loading the data.
    replace = {"billion US$2005/yr": "billion USD_2005/yr"}

    def __init__(self, source, source_kw):
        s = "ICONICS:SSP(2017)."
        if not source.startswith(s):
            raise ValueError(source)

        *parts, self.ssp_number = source.partition(s)

        # Map the `measure` keyword to a string appearing in the data
        _kw = copy(source_kw)
        self.measure = {
            "GDP": "GDP|PPP",
            "POP": "Population",
        }[_kw.pop("measure")]

        # Store the model ID, if any
        self.model = _kw.pop("model", None)

        # Determine the date based on the model ID. There is a 1:1 correspondence.
        self.date = self.model_date[self.model]

        if len(_kw):
            raise ValueError(_kw)

    def __call__(self):
        # Assemble a query string
        extra = "d" if self.ssp_number == "4" and self.model == "IIASA-WiC POP" else ""
        query = " and ".join(
            [
                f"SCENARIO == 'SSP{self.ssp_number}{extra}_v9_{self.date}'",
                f"VARIABLE == '{self.measure}'",
                f"MODEL == '{self.model}'" if self.model else "True",
            ]
        )
        log.debug(query)

        parts = ("ssp", "SspDb_country_data_2013-06-12.csv.zip")
        if HAS_MESSAGE_DATA:
            path = private_data_path(*parts)
        else:
            path = package_data_path("test", *parts)
            log.warning(f"Reading random data from {path}")

        return iamc_like_data_for_query(path, query, replace=self.replace)


@register_source
class SSPUpdate(ExoDataSource):
    """Provider of exogenous data from the SSP Update database.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    arguments:

    - `source`: Any value from :data:`.SSP_2024` or equivalent string, for instance
      "ICONICS:SSP(2024).2".

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="ICONICS:SSP(2024).3",
    ...     source_kw=dict(measure="GDP", model="IIASA GDP 2023"),
    ... )
    >>> result = computer.get(keys[0])
    """

    id = "SSP update"

    def __init__(self, source, source_kw):
        s = "ICONICS:SSP(2024)."
        if not source.startswith(s):
            raise ValueError(source)

        *parts, self.ssp_number = source.partition(s)

        # Map the `measure` keyword to a string appearing in the data
        _kw = copy(source_kw)
        self.measure = {
            "GDP": "GDP|PPP",
            "POP": "Population",
        }[_kw.pop("measure")]

        # Store the model ID, if any
        self.model = _kw.pop("model", None)

        if len(_kw):
            raise ValueError(_kw)

    def __call__(self):
        # Assemble a query string
        query = " and ".join(
            [
                f"Scenario == 'SSP{self.ssp_number} - Review Phase 1'",
                f"Variable == '{self.measure}'",
                f"Model == '{self.model}'" if self.model else "True",
            ]
        )

        parts = ("ssp", "SSP-Review-Phase-1.csv.gz")
        if HAS_MESSAGE_DATA:
            path = private_data_path(*parts)
        else:
            path = package_data_path("test", *parts)
            log.warning(f"Reading random data from {path}")

        return iamc_like_data_for_query(path, query)



import logging
from copy import deepcopy

import pytest
import sdmx.model.v21 as sdmx_model
import xarray as xr
from genno.caching import hash_args
from ixmp.testing import assert_logs

import message_ix_models.util.cache
from message_ix_models import ScenarioInfo
from message_ix_models.util import cached

log = logging.getLogger(__name__)


class TestEncoder:
    def test_sdmx(self):
        """:mod:`message_ix_models` configures :class:`.Encoder` for |Code|."""
        codes0 = [sdmx_model.Code(id=f"FOO{i}", name="foo") for i in range(5)]
        codes1 = [f"FOO{i}" for i in range(5)]

        # List of codes hashes the same as a list of their string IDs
        expected = "40a0735385448dcbe745904ebfec7255995ca451"
        assert expected == hash_args(codes0, bar="baz") == hash_args(codes1, bar="baz")


def test_cached(caplog, test_context, tmp_path):
    """:func:`.cached` works as expected.

    .. todo:: test behaviour when :data:`.SKIP_CACHE` is :obj:`True`
    """
    # Clear seen paths, so that log message below is guaranteed to occur
    message_ix_models.util.cache.PATHS_SEEN.clear()

    # Store in the temporary directory for this session, to avoid collisions across
    # sessions
    test_context.cache_path = tmp_path.joinpath("cache")

    # A dummy path to be hashed as an argument
    path_foo = tmp_path.joinpath("foo", "bar")

    with caplog.at_level(logging.DEBUG, logger="message_ix_models"):

        @cached
        def func0(ctx, a, path, b=3):
            """A test function."""
            log.info("func0 runs")
            return f"{id(ctx)}, {a + b}"

    # Docstring is modified
    assert "Data returned by this function is cached" in func0.__doc__
    # Message is logged
    assert f"func0() will cache in {tmp_path.joinpath('cache')}" in caplog.messages

    @cached
    def func1(x=1, y=2, **kwargs):
        # Function with defaults for all arguments
        log.info("func1 runs")
        return x + y

    caplog.clear()

    # pathlib.Path argument is serialized to JSON as part of the argument hash;
    # function runs, messages logged
    with assert_logs(caplog, "func0 runs"):
        result0 = func0(test_context, 1, path_foo)

    caplog.clear()
    result1 = func0(test_context, 1, path_foo)
    # Function does not run
    assert "func0 runs" not in caplog.messages
    assert caplog.messages[0].startswith("Cache hit for func0")
    # Results identical
    assert result0 == result1

    # Different context object with identical contents hashes equal
    ctx2 = deepcopy(test_context)
    assert id(test_context) != id(ctx2)

    result2 = func0(ctx2, 1, path_foo)
    # Function does not run
    assert "func0 runs" not in caplog.messages
    # Results are identical, i.e. including the old ID
    assert result0 == result2

    ctx2.delete()
    caplog.clear()

    # Hash of no arguments is the same, function only runs once
    assert 3 == func1() == func1()
    assert 1 == sum(m == "func1 runs" for m in caplog.messages)

    # Warnings logged for unhashables; ScenarioInfo is hashed as dict
    caplog.clear()
    with assert_logs(
        caplog,
        [
            "ignores <class 'xarray.core.dataset.Dataset'>",
            "ignores <class 'ixmp.core.platform.Platform'>",
        ],
    ):
        func1(ds=xr.Dataset(), mp=test_context.get_platform(), si=ScenarioInfo())

    # Unserializable type raises an exception
    with pytest.raises(
        TypeError, match="Object of type slice is not JSON serializable"
    ):
        func1(arg=slice(None))



import sys
from importlib.metadata import PackageNotFoundError, version

import pint
from iam_units import registry

from message_ix_models.util._logging import setup as setup_logging
from message_ix_models.util.config import Config
from message_ix_models.util.context import Context
from message_ix_models.util.importlib import MessageDataFinder
from message_ix_models.util.scenarioinfo import ScenarioInfo, Spec
from message_ix_models.workflow import Workflow

# Expose utility classes
__all__ = ["Config", "Context", "ScenarioInfo", "Spec", "Workflow"]

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    # Package is not installed
    __version__ = "999"

# No logging to stdout (console) by default
setup_logging(console=False)

# Use iam_units.registry as the default pint.UnitsRegistry
pint.set_application_registry(registry)

# Ensure at least one Context instance is created
Context()

# Use this finder only if others fail
sys.meta_path.append(MessageDataFinder())



"""Command-line utilities.

These are used for building CLIs using :mod:`click`.
"""
import logging
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Optional, Union

import click
from click import Argument, Choice, Option

from message_ix_models import Context, model
from message_ix_models.model.structure import codelists

from .scenarioinfo import ScenarioInfo

log = logging.getLogger(__name__)


def common_params(param_names: str):
    """Decorate a click.command with common parameters `param_names`.

    `param_names` must be a space-separated string of names appearing in :data:`PARAMS`,
    e.g. ``"ssp force output_model"``. The decorated function receives keyword
    arguments with these names::

        @click.command()
        @common_params("ssp force output_model")
        def mycmd(ssp, force, output_model)
            # ...
    """

    # Simplified from click.decorators._param_memo
    def decorator(f):
        if not hasattr(f, "__click_params__"):
            f.__click_params__ = []
        f.__click_params__.extend(
            PARAMS[name] for name in reversed(param_names.split())
        )
        return f

    return decorator


def default_path_cb(*default_parts):
    """Return a callback function for click.Option handling.

    If no option value is given, the callback uses :meth:`Context.get_local_path` and
    `default_parts` to provide a path that is relative to local data directory, e.g.
    the current working directory (see :doc:`/data`).
    """

    def _callback(context, param, value):
        value = value or context.obj.get_local_path(*default_parts)
        setattr(context.obj, param.name, value)
        return value

    return _callback


def exec_cb(expression: str) -> Callable:
    """Return a callback that :func:`exec`-utes an `expression`.

    The `expression` is executed in a limited context that has only two names available:

    - :py:`context`: the :class:`.Context` instance.
    - :py:`value`: the value passed to the :mod:`click.Parameter`.

    Example
    -------
    >>> @click.command
    ... @click.option(
    ...     "--myopt", callback=exec_cb("context.my_mod.my_opt = value + 3")
    ... )
    ... def cmd(...):
    ...     ...
    """

    def _cb(context: Union[click.Context, Context], param, value):
        ctx = context.obj if isinstance(context, click.Context) else context
        exec(expression, {}, {"context": ctx, "value": value})
        return value

    return _cb


def format_sys_argv() -> str:
    """Format :data:`sys.argv` in a readable manner."""
    import sys

    lines = ["Invoked:"]
    indent = ""
    for item in sys.argv:
        lines.append(f"{indent}{item} \\")
        indent = "  "

    return "\n".join(lines)[:-2]


def store_context(context: Union[click.Context, Context], param, value):
    """Callback that simply stores a value on the :class:`.Context` object.

    Use this for parameters that are not used directly in a @click.command() function,
    but need to be carried by the Context for later use.
    """
    setattr(
        context.obj if isinstance(context, click.Context) else context,
        param.name,
        value,
    )
    return value


def urls_from_file(
    context: Union[click.Context, Context], param, value
) -> List[ScenarioInfo]:
    """Callback to parse scenario URLs from `value`."""
    si: List[ScenarioInfo] = []

    if value is None:
        return si

    with click.open_file(value) as f:
        for line in f:
            si.append(ScenarioInfo.from_url(url=line))

    # Store on context
    mm_context = context.obj if isinstance(context, click.Context) else context
    mm_context.core.scenarios = si

    return si


def unique_id() -> str:
    """Return a unique ID for a CLI invocation.

    The return value resembles "mix-models-debug-3332d415ef65bf2a-2023-02-02T162931" and
    contains:

    - The CLI name and (sub)commands.
    - A hash of all the CLI parameters (options and arguments).
    - The current date and time, in ISO format with Windows-incompatible ":" removed.

    """
    click_context = click.get_current_context()

    # Collapse CLI (sub)commands and their arguments to a hashable data structure
    # This also includes CLI options *not* given
    c: Optional[click.Context] = click_context
    data = []
    while c is not None:
        data.append((c.command.name, tuple(sorted(c.params.items()))))
        c = c.parent

    # Assemble parts
    return "-".join(
        [
            click_context.command_path.replace(" ", "-"),
            f"{hash(tuple(reversed(data))):x}",
            datetime.now().isoformat(timespec="seconds").replace(":", ""),
        ]
    )


#: Common command-line parameters (arguments and options). See :func:`common_params`.
PARAMS = {
    "dest": Option(
        ["--dest"],
        callback=store_context,
        expose_value=False,
        help="Destination URL for created scenario(s).",
    ),
    "dry_run": Option(
        ["--dry-run"],
        is_flag=True,
        callback=store_context,
        expose_value=False,
        help="Only show what would be done.",
    ),
    "force": Option(
        ["--force"],
        is_flag=True,
        callback=store_context,
        help="Overwrite or modify existing model/scenario.",
    ),
    "nodes": Option(
        ["--nodes"],
        help="Code list to use for 'node' dimension.",
        callback=exec_cb("context.model.regions = value"),
        type=Choice(codelists("node")),
        default=model.Config.regions,
        expose_value=False,
    ),
    "output_model": Option(
        ["--output-model"], help="Model name under which scenarios should be generated."
    ),
    "policy_path": Option(
        ["--policy-path"],
        callback=default_path_cb("scenario_generation", "policies"),
        help="Path to policy scripts.",
    ),
    "platform_dest": Option(["--platform-dest"], help="Name of destination Platform."),
    "quiet": Option(
        ["--quiet"],
        is_flag=True,
        expose_value=False,
        help="Show less or no output.",
    ),
    "regions": Option(
        ["--regions"],
        help="Code list to use for 'node' dimension.",
        callback=exec_cb("context.model.regions = value or context.model.regions"),
        type=Choice(codelists("node")),
    ),
    "rep_out_path": Option(
        ["--rep-out-path"],
        callback=default_path_cb("reporting_output"),
        help="Path for reporting output.",
    ),
    "rep_template": Option(
        ["--rep-template"],
        callback=default_path_cb(
            "message_data", "tools", "post_processing", "MESSAGEix_WorkDB_Template.xlsx"
        ),
        help="Path incl. filename and extension to reporting template.",
    ),
    "run_reporting_only": Option(
        ["--run-reporting-only"],
        is_flag=True,
        callback=store_context,
        help="Run only reporting.",
    ),
    "ssp": Argument(
        ["ssp"], callback=store_context, type=Choice(["SSP1", "SSP2", "SSP3"])
    ),
    "urls_from_file": Option(
        ["--urls-from-file", "-f"],
        type=click.Path(
            exists=True,
            dir_okay=False,
            resolve_path=True,
            allow_dash=True,
            path_type=Path,
        ),
        callback=urls_from_file,
    ),
    "verbose": Option(
        # NB cannot use store_callback here; this is processed in the top-level CLI
        #    before the message_ix_models.Context() object is set up
        ["--verbose", "-v"],
        is_flag=True,
        help="Print DEBUG-level log messages.",
    ),
    "years": Option(
        ["--years"],
        help="Code list to use for the 'year' dimension.",
        callback=exec_cb("context.model.years = value"),
        type=Choice(codelists("year")),
        default=model.Config.years,
        # expose_value=False,
    ),
}



import logging
from contextlib import nullcontext
from copy import deepcopy
from functools import partial
from importlib import import_module
from operator import itemgetter
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Union
from warnings import warn

import genno.config
import yaml
from genno import Key, KeyExistsError
from genno.compat.pyam import iamc as handle_iamc
from message_ix import Reporter, Scenario

from message_ix_models import Context, ScenarioInfo
from message_ix_models.util import minimum_version
from message_ix_models.util._logging import mark_time, silence_log

from .config import Config

__all__ = [
    "Config",
    "prepare_reporter",
    "register",
    "report",
]


log = logging.getLogger(__name__)

# Add to the configuration keys stored by Reporter.configure().
genno.config.STORE.add("output_path")
genno.config.STORE.add("output_dir")

#: List of callbacks for preparing the Reporter.
CALLBACKS: List[Callable] = []


# Ignore a section in global.yaml used to define YAML anchors
@genno.config.handles("_iamc formats")
def _(c: Reporter, info):
    pass


@genno.config.handles("iamc")
def iamc(c: Reporter, info):
    """Handle one entry from the ``iamc:`` config section.

    This version overrides the version from :mod:`genno.config` to:

    - Set some defaults for the `rename` argument for :meth:`.convert_pyam`:

      - The `n` and `nl` dimensions are mapped to the "region" IAMC column.
      - The `y`, `ya`, and `yv` dimensions are mapped to the "year" column.

    - Use the MESSAGEix-GLOBIOM custom :func:`.util.collapse` callback to perform
      renaming etc. while collapsing dimensions to the IAMC ones. The "var" key from
      the entry, if any, is passed to the `var` argument of that function.

    - Provide optional partial sums. The "sums" key of the entry can give a list of
      strings such as ``["x", "y", "x-y"]``; in this case, the conversion to IAMC format
      is also applied to the same "base" key with a partial sum over the dimension "x";
      over "y", and over both "x" and "y". The corresponding dimensions are omitted from
      "var". All data are concatenated.
    """
    # FIXME the upstream key "variable" for the configuration is confusing; choose a
    #       better name
    from message_ix_models.report.util import collapse

    # Common
    base_key = Key(info["base"])

    # Use message_ix_models custom collapse() method
    info.setdefault("collapse", {})

    # Add standard renames
    info.setdefault("rename", {})
    for dim, target in (
        ("n", "region"),
        ("nl", "region"),
        ("y", "year"),
        ("ya", "year"),
        ("yv", "year"),
    ):
        info["rename"].setdefault(dim, target)

    # Iterate over partial sums
    # TODO move some or all of this logic upstream
    keys = []  # Resulting keys
    for dims in [""] + info.pop("sums", []):
        # Dimensions to partial
        # TODO allow iterable of str
        dims = dims.split("-")

        label = f"{info['variable']} {'-'.join(dims) or 'full'}"

        # Modified copy of `info` for this invocation
        _info = info.copy()
        # Base key: use the partial sum over any `dims`. Use a distinct variable name.
        _info.update(base=base_key.drop(*dims), variable=label)
        # Exclude any summed dimensions from the IAMC Variable to be constructed
        _info["collapse"].update(
            callback=partial(
                collapse, var=list(filter(lambda v: v not in dims, info.get("var", [])))
            )
        )

        # Invoke the genno built-in handler
        handle_iamc(c, _info)

        keys.append(f"{label}::iamc")

    # Concatenate together the multiple tables
    c.add("concat", f"{info['variable']}::iamc", *keys)


def register(name_or_callback: Union[Callable, str]) -> Optional[str]:
    """Register a callback function for :meth:`prepare_reporter`.

    Each registered function is called by :meth:`prepare_reporter`, in order to add or
    modify reporting keys. Specific model variants and projects can register a callback
    to extend the reporting graph.

    Callback functions must take two arguments: the Reporter, and a :class:`.Context`:

    .. code-block:: python

        from message_ix.report import Reporter
        from message_ix_models import Context
        from message_ix_models.report import register

        def cb(rep: Reporter, ctx: Context):
            # Modify `rep` by calling its methods ...
            pass

        register(cb)

    Parameters
    ----------
    name_or_callback
        If a string, this may be a submodule of :mod:`.message_ix_models`, or
        :mod:`message_data`, in which case the function
        ``{message_data,message_ix_models}.{name}.report.callback`` is used. Or, it may
        be a fully-resolved package/module name, in which case ``{name}.callback`` is
        used. If a callable (function), it is used directly.
    """
    if isinstance(name_or_callback, str):
        # Resolve a string
        candidates = [
            # As a fully-resolved package/module name
            name_or_callback,
            # As a submodule of message_ix_models
            f"message_ix_models.{name_or_callback}.report",
            # As a submodule of message_data
            f"message_data.{name_or_callback}.report",
        ]
        mod = None
        for name in candidates:
            try:
                mod = import_module(name)
            except ModuleNotFoundError:
                continue
            else:
                break
        if mod is None:
            raise ModuleNotFoundError(" or ".join(candidates))
        callback = mod.callback
    else:
        callback = name_or_callback
        name = callback.__name__

    if callback in CALLBACKS:
        log.info(f"Already registered: {callback}")
        return None

    CALLBACKS.append(callback)
    return name


def log_before(context, rep, key) -> None:
    log.info(f"Prepare to report {'(DRY RUN)' if context.dry_run else ''}")
    log.info(key)
    log.log(
        logging.INFO
        if (context.core.dry_run or context.core.verbose)
        else logging.DEBUG,
        "\n" + rep.describe(key),
    )
    mark_time()


def report(context: Context, *args, **kwargs):
    """Report (post-process) solution data in a |Scenario| and store time series data.

    This function provides a single, common interface to call both the :mod:`genno`
    -based (:mod:`message_ix_models.report`) and ‘legacy’ (
    :mod:`message_data.tools.post_processing`) reporting codes.

    Parameters
    ----------
    context : Context
        The code responds to:

        - :attr:`.dry_run`: if :obj:`True`, reporting is prepared but nothing is done.
        - :attr:`~.Config.scenario_info` and :attr:`~.Config.platform_info`: used to
          retrieve the Scenario to be reported.

        - :py:`context.report`, which is an instance of :class:`.report.Config`; see
          there for available configuration settings.
    """
    from message_ix_models.util.ixmp import discard_on_error

    # Handle deprecated usage that appears in:
    # - .model.cli.new_baseline()
    # - .model.create.solve()
    # - .projects.covid.scenario_runner.ScenarioRunner.solve()
    if isinstance(context, Scenario):
        warn(
            "Calling report(scenario, path, legacy=…); pass a Context instead",
            category=DeprecationWarning,
        )
        # Ensure `context` is actually a Context object for the following code
        scenario = context
        context = Context.get_instance(-1)

        # Transfer args, kwargs to context
        context.set_scenario(scenario)
        context.report.legacy.update(kwargs.pop("legacy", {}))

        if len(args) + len(set(kwargs.keys()) & {"path"}) != 1:
            raise TypeError(
                f"Unknown mix of deprecated positional {args!r} "
                f"and keyword arguments {kwargs!r}"
            )
        elif len(args) == 1:
            out_dir = args[0]
        else:
            out_dir = kwargs.pop("path")
        context.report.legacy.setdefault("out_dir", out_dir)

    if context.report.legacy["use"]:
        return _invoke_legacy_reporting(context)

    with (
        nullcontext()
        if context.core.verbose
        else silence_log(["genno", "message_ix_models"])
    ):
        rep, key = prepare_reporter(context)

    log_before(context, rep, key)

    if context.dry_run:
        return

    with discard_on_error(rep.graph["scenario"]):
        result = rep.get(key)

    # Display information about the result
    log.info(f"Result:\n\n{result}\n")
    log.info(
        f"File output(s), if any, written under:\n{rep.graph['config']['output_dir']}"
    )


def _invoke_legacy_reporting(context):
    log.info("Using tools.post_processing.iamc_report_hackathon")
    from message_data.tools.post_processing import iamc_report_hackathon

    # Convert "legacy" config to keyword arguments for .iamc_report_hackathon.report()
    kwargs = deepcopy(context.report.legacy)
    kwargs.pop("use")

    # Read a legacy reporting configuration file and update the arguments
    config_file_path = kwargs.pop("config_file_path", None)
    if isinstance(config_file_path, Path) and config_file_path.exists():
        with open(config_file_path, "r") as f:
            kwargs.update(yaml.safe_load(f))

    # Retrieve the Scenario and Platform
    scen = context.get_scenario()
    mp = scen.platform

    mark_time()

    # `context` is passed only for the "dry_run" setting; the function receives all its
    # other settings via the `kwargs`
    return iamc_report_hackathon.report(mp=mp, scen=scen, context=context, **kwargs)


@minimum_version("message_ix 3.6")
def prepare_reporter(
    context: Context,
    scenario: Optional[Scenario] = None,
    reporter: Optional[Reporter] = None,
) -> Tuple[Reporter, Key]:
    """Return a :class:`message_ix.Reporter` and `key` prepared to report a |Scenario|.

    Parameters
    ----------
    context : Context
        The code responds to :py:`context.report`, which is an instance of
        :class:`.report.Config`.
    scenario : Scenario, *optional*
        Scenario to report. If not given, :meth:`.Context.get_scenario` is used to
        retrieve a Scenario.
    reporter : message_ix.Reporter, *optional*
        Existing reporter to extend with computations. If not given, it is created
        using :meth:`message_ix.Reporter.from_scenario`.

    Returns
    -------
    .Reporter
        Reporter prepared with MESSAGEix-GLOBIOM calculations; if `reporter` is given,
        this is a reference to the same object.

        If :attr:`.cli_output` is given, a task with the key "cli-output" is added that
        writes the :attr:`.Config.key` to that path.
    .Key
        Same as :attr:`.Config.key` if any, but in full resolution; else either
        "default" or "cli-output" according to the other settings.
    """
    log.info("Prepare reporter")

    if reporter:
        # Existing `Reporter` provided
        rep = reporter
        has_solution = True
        if scenario:
            log.warning(f"{scenario = } argument ignored")
        scenario = rep.graph["scenario"]
    else:
        # Retrieve the scenario
        scenario = scenario or context.get_scenario()
        # Create a new Reporter
        rep = Reporter.from_scenario(scenario)
        has_solution = scenario.has_solution()

    # Append the message_data operators
    rep.require_compat("message_ix_models.report.operator")

    # Force re-installation of the function iamc() in this file as the handler for
    # "iamc:" sections in global.yaml. Until message_data.reporting is removed, then
    # importing it will cause the iamc() function in *that* file to override the one
    # registered above.
    # TODO Remove, once message_data.reporting is removed.
    genno.config.handles("iamc")(iamc)

    if context.report.use_scenario_path:
        # Construct ScenarioInfo
        si = ScenarioInfo(scenario, empty=True)
        # Use the scenario URL to extend the path
        context.report.set_output_dir(context.report.output_dir.joinpath(si.path))

    # Pass values to genno's configuration; deepcopy to protect from destructive
    # operations
    rep.configure(
        **deepcopy(context.report.genno_config),
        fail="raise" if has_solution else logging.NOTSET,
    )
    rep.configure(model=deepcopy(context.model))

    # Apply callbacks for other modules which define additional reporting computations
    for callback in CALLBACKS:
        callback(rep, context)

    key = context.report.key
    if key:
        # If just a bare name like "ACT" is given, infer the full key
        if Key.bare_name(key):
            inferred = rep.infer_keys(key)
            if inferred != key:
                log.info(f"Infer {inferred!r} for {key!r}")
                key = inferred

        if context.report.cli_output:
            # Add a new task that writes `key` to the specified file
            key = rep.add(
                "cli-output", "write_report", key, path=context.report.cli_output
            )
    else:
        key = rep.default_key
        log.info(f"No key given; will use default: {key!r}")

    # Create the output directory
    context.report.mkdir()

    log.info("…done")

    return rep, key


def defaults(rep: Reporter, context: Context) -> None:
    from message_ix_models.model.structure import get_codes

    from .util import add_replacements

    # Add mappings for conversions to IAMC data structures
    add_replacements("c", get_codes("commodity"))
    add_replacements("t", get_codes("technology"))

    # Ensure "y::model" and "y0" are present
    # TODO remove this once message-ix-models depends on message_ix > 3.7.0 at minimum
    for comp in (
        ("y::model", "model_periods", "y", "cat_year"),
        ("y0", itemgetter(0), "y::model"),
    ):
        try:
            rep.add(*comp, strict=True)
        except KeyExistsError:
            pass  # message_ix > 3.7.0; these are already defined


register(defaults)
register("message_ix_models.report.plot")



"""Prepare base models from snapshot data."""
import logging
from pathlib import Path

import pandas as pd
from message_ix import Scenario
from message_ix.models import MACRO
from tqdm import tqdm

from message_ix_models import Spec
from message_ix_models.util import minimum_version
from message_ix_models.util.pooch import SOURCE, fetch

from .build import apply_spec
from .structure import get_codes

log = logging.getLogger(__name__)


def unpack(path: Path) -> Path:
    """Unpack :ref:`ixmp-format Excel file <ixmp:excel-data-format>` at `path`.

    The file is unpacked into a directory with the same name stem as the file (that is,
    without the :file:`.xlsx` suffix). In this directory are created:

    - One :file:`.csv.gz` file for each MESSAGE and/or MACRO parameter.
    - One file :file:`sets.xlsx` with only the :mod:`ixmp` sets, and no parameter data.

    If the files exist, they are not updated.
    To force re-unpacking, delete the files.

    Returns
    -------
    Path
        Path to the directory containing the unpacked files.
    """
    assert path.suffix == ".xlsx"
    base = path.with_suffix("")
    base.mkdir(exist_ok=True)

    # Get item name -> ixmp type mapping as a pd.Series
    xf = pd.ExcelFile(path, engine="openpyxl")
    name_type = xf.parse("ix_type_mapping")

    # Copied exactly from ixmp.backend.io
    def parse_item_sheets(name):
        """Read data for item *name*, possibly across multiple sheets."""
        dfs = [xf.parse(name)]

        # Collect data from repeated sheets due to max_row limit
        for x in filter(lambda n: n.startswith(name + "("), xf.sheet_names):
            dfs.append(xf.parse(x))  # pragma: no cover

        # Concatenate once and return
        return pd.concat(dfs, axis=0, ignore_index=True)

    sets_path = base.joinpath("sets.xlsx")
    sets_path.unlink(missing_ok=True)

    with pd.ExcelWriter(sets_path, engine="openpyxl") as ew:
        for _, (name, ix_type) in tqdm(name_type.iterrows()):
            item_path = base.joinpath(f"{name}.csv.gz")
            if item_path.exists():
                continue

            df = parse_item_sheets(name)

            if ix_type == "set":
                df.to_excel(ew, sheet_name=name, index=False)
            else:  # pragma: no cover
                df.to_csv(item_path, index=False)

        name_type.query("ix_type == 'set'").to_excel(ew, sheet_name="ix_type_mapping")

    return base


def read_excel(scenario: Scenario, path: Path) -> None:
    """Similar to :meth:`.Scenario.read_excel`, but using :func:`unpack`."""
    base = unpack(path)

    scenario.read_excel(base.joinpath("sets.xlsx"))

    parameters = set(scenario.par_list())

    with scenario.transact(f"Read snapshot data from {path}"):
        for p in base.glob("*.csv.gz"):
            name = p.name.split(".")[0]

            if name not in parameters:
                continue  # Variable or equation data: don't read

            data = pd.read_csv(p)

            # Correct units
            if name == "inv_cost":
                data.replace({"unit": {"USD_2005/t ": "USD_2005/t"}}, inplace=True)

            scenario.add_par(name, data)


@minimum_version("message_ix 3.5")
def load(scenario: Scenario, snapshot_id: int) -> None:
    """Fetch and load snapshot with ID `snapshot_id` into `scenario`.

    See also
    --------
    SNAPSHOTS
    """
    path, *_ = fetch(**SOURCE[f"snapshot-{snapshot_id}"])

    # Add units
    spec = Spec()
    spec.add.set["unit"] = get_codes("unit/snapshot")
    apply_spec(scenario, spec)

    # Initialize MACRO items
    with scenario.transact("Prepare scenario for snapshot data"):
        MACRO.initialize(scenario)

    read_excel(scenario, path)



from dataclasses import dataclass, field, fields

from message_ix_models.model.structure import codelists
from message_ix_models.util.context import _ALIAS


@dataclass
class Config:
    """Settings and valid values for :mod:`message_ix_models.model` and submodules.

    For backwards compatibility, it is possible to access these on a :class:`Context`
    using:

    .. code-block:: python

       c = Context()
       c.regions = "R14"

    …however, it is best to access them explicitly as:

    .. code-block:: python

       c.model.regions = "R14"
    """

    #: The 'node' codelist (regional aggregation) to use. Must be one of the lists of
    #: nodes described at :doc:`/pkg-data/node`.
    regions: str = "R14"

    #: The 'relations' codelist to use. Must be one of the lists of relations described
    #: at :doc:`/pkg-data/relation`.
    relations: str = "A"

    #: The 'year' codelist (time periods) to use, Must be one of the lists of periods
    #: described at :doc:`/pkg-data/year`.
    years: str = "B"

    #: Create the reference energy system with dummy commodities and technologies. See
    #: :func:`.bare.get_spec`.
    res_with_dummies: bool = False

    #: Default or preferred units for model quantities and reporting.
    units: dict = field(
        default_factory=lambda: {"energy": "GWa", "CO2 emissions": "Mt / a"}
    )

    def check(self):
        """Check the validity of :attr:`regions`, :attr:`relations`, :attr:`years`."""
        for attr, name in [
            ("regions", "node"),
            ("relations", "relation"),
            ("years", "year"),
        ]:
            if getattr(self, attr) not in codelists(name):
                raise ValueError(
                    f"{attr}={getattr(self, attr)!r} not among {codelists(name)}"
                )


# Extend the list of settings that can be set directly on a Context instance.
_ALIAS.update({f.name: "model" for f in fields(Config)})



import logging

log = logging.getLogger(__name__)


def get_data(scenario, context, spec, **options):
    """Data for the bare RES."""
    if context.model.res_with_dummies:
        log.warning("get_dummy_data() not migrated")
        # return get_dummy_data(scenario, spec)
    else:
        return dict()



"""Logging utilities."""

import logging
import logging.config
from contextlib import contextmanager
from copy import deepcopy
from time import process_time

__all__ = [
    "Formatter",
    "make_formatter",
    "setup",
    "silence_log",
]

log = logging.getLogger(__name__)


@contextmanager
def silence_log(names=None, level=logging.ERROR):
    """Context manager to temporarily quiet 1 or more loggers.

    Parameters
    ----------
    names : str, *optional*
        Space-separated names of loggers to quiet.
    level : int, *optional*
        Minimum level of log messages to allow.

    Examples
    --------
    >>> with silence_log():
    >>>     log.warning("This message is not recorded.")
    """
    # Default: the top-level logger for the package containing this file
    if names is None:
        names = [__name__.split(".")[0], "message_data"]
    elif isinstance(names, str):
        names = [names]

    log.info(f"Set level={level} for logger(s): {' '.join(names)}")

    # Retrieve the logger objects
    loggers = list(map(logging.getLogger, names))
    # Store their current levels
    levels = []

    try:
        for logger in loggers:
            levels.append(logger.getEffectiveLevel())  # Store the current levels
            logger.setLevel(level)  # Set the level
        yield
    finally:
        # Restore the levels
        for logger, original_level in zip(loggers, levels):
            logger.setLevel(original_level)
        log.info("…restored.")


@contextmanager
def preserve_log_level():
    """Context manager to preserve the level of the ``message_ix_models`` logger."""
    # Get the top-level logger for the package containing this file
    main_log = logging.getLogger(__name__.split(".")[0])

    try:
        # Store the current level
        level = main_log.getEffectiveLevel()
        yield
    finally:
        # Restore the level
        main_log.setLevel(level)


class Formatter(logging.Formatter):
    """Formatter for log records.

    Parameters
    ----------
    colorama : module
        If provided, :mod:`colorama` is used to colour log messages printed to stdout.
    """

    CYAN = ""
    DIM = ""
    RESET_ALL = ""

    _short_name = None

    def __init__(self, colorama):
        super().__init__()
        if colorama:
            self.CYAN = colorama.Fore.CYAN
            self.DIM = colorama.Style.DIM
            self.RESET_ALL = colorama.Style.RESET_ALL

    def format(self, record):
        """Format `record`.

        Records are formatted like::

            model.transport.data.add_par_data  220 rows in 'input'
            ...add_par_data:  further messages

        …with the calling function name (e.g. 'add_par_data') coloured for legibility
        on first occurrence, then dimmed when repeated.
        """
        # Remove the leading 'message_data.' from the module name
        name_parts = record.name.split(".")
        if name_parts[0] in ("message_ix_models", "message_data"):
            short_name = ".".join(["—"] + name_parts[1:])
        else:
            short_name = record.name

        if self._short_name != short_name:
            self._short_name = short_name
            prefix = f"{self.CYAN}{short_name}."
        else:
            prefix = f"{self.DIM}..."

        return f"{prefix}{record.funcName}{self.RESET_ALL}  {record.getMessage()}"


def make_formatter():
    """Return a :class:`Formatter` instance for the ``message_ix_models`` logger.

    See also
    --------
    setup
    """
    try:
        # Initialize colorama
        import colorama

        colorama.init()
    except ImportError:  # pragma: no cover
        # Colorama not installed
        colorama = None

    return Formatter(colorama)


_TIMES = []


def mark_time(quiet=False):
    """Record and log (if `quiet` is :obj:`True`) a time mark."""
    _TIMES.append(process_time())
    if not quiet and len(_TIMES) > 1:
        logging.getLogger(__name__).info(
            f" +{_TIMES[-1] - _TIMES[-2]:.1f} = {_TIMES[-1]:.1f} seconds"
        )


CONFIG = dict(
    version=1,
    disable_existing_loggers=False,
    formatters=dict(simple={"()": "message_ix_models.util._logging.make_formatter"}),
    handlers=dict(
        console={
            "class": "logging.StreamHandler",
            "level": "NOTSET",
            "formatter": "simple",
            "stream": "ext://sys.stdout",
        },
        # commented: needs code in setup() to choose an appropriate file path
        # file_handler={
        #    "class": "logging.handlers.RotatingFileHandler",
        #    "level": "DEBUG",
        #    "formatter": "simple",
        #    "backupCount": "100",
        #    "delay": True,
        # },
    ),
    loggers=dict(
        message_ix_models=dict(
            level="NOTSET",
            # propagate=False,
            # handlers=[],
        ),
        message_data=dict(
            level="NOTSET",
            # propagate=False,
            # handlers=[],
        ),
    ),
    root=dict(
        handlers=[],
    ),
)


def setup(
    level="NOTSET",
    console=True,
    # file=False,
):
    """Initialize logging.

    Parameters
    ----------
    level : str, *optional*
        Log level for :mod:`message_ix_models` and :mod:`message_data`.
    console : bool, *optional*
        If :obj:`True`, print all messages to console using a :class:`Formatter`.
    """
    # Copy to avoid modifying with the operations below
    config = deepcopy(CONFIG)

    config["root"].setdefault("level", level)

    if console:
        config["root"]["handlers"].append("console")
    # if file:
    #     config["loggers"]["message_data"]["handlers"].append("file")

    # Apply the configuration
    logging.config.dictConfig(config)



"""Manipulate data structures for working with the SSPs."""
import logging
from textwrap import wrap
from typing import TYPE_CHECKING, Optional

import sdmx
import sdmx.model.v30 as m
import sdmx.urn

from message_ix_models.util.sdmx import make_enum, register_agency, write

if TYPE_CHECKING:
    from os import PathLike

    from message_ix_models import Context

log = logging.getLogger(__name__)


DESC = """
This code list is *not* officially published by ICONICS; rather, it is a rendering into
SDMX of structural information that is provided by ICONICS participants. It may change
or be superseded at any time.

Each code has a single digit ID like "1"; the "original-id" annotation (and the start of
the name) give a string like "SSP1". These original IDs are *only* unique if using data
enumerated by solely by one code list or the other; if mixing the two, then they will be
ambiguous. The URNs of the codes or parts thereof (for instance, "ICONICS:SSP(2017).1")
are, by construction, unique.
"""

CL_INFO = (
    dict(
        version="2017",
        name_extra="2017 edition",
        desc_extra="This is the original set of SSPs as described in"
        " https://doi.org/10.1016/j.gloenvcha.2016.05.009.",
    ),
    dict(
        version="2024",
        name_extra="2024 edition",
        desc_extra="This set of SSPs is currently under development; for details, see "
        "https://depts.washington.edu/iconics/.",
    ),
)

CODE_INFO = (
    dict(
        id="1",
        name="Sustainability",
        name_long="Sustainability – Taking the Green Road",
        description="""Low challenges to mitigation and adaptation.

The world shifts gradually, but pervasively, toward a more sustainable path, emphasizing
more inclusive development that respects perceived environmental boundaries. Management
of the global commons slowly improves, educational and health investments accelerate the
demographic transition, and the emphasis on economic growth shifts toward a broader
emphasis on human well-being. Driven by an increasing commitment to achieving
development goals, inequality is reduced both across and within countries. Consumption
is oriented toward low material growth and lower resource and energy intensity.""",
    ),
    dict(
        id="2",
        name="Middle of the Road",
        name_long="Middle of the Road",
        description="""Medium challenges to mitigation and adaptation.

The world follows a path in which social, economic, and technological trends do not
shift markedly from historical patterns. Development and income growth proceeds
unevenly, with some countries making relatively good progress while others fall short of
expectations. Global and national institutions work toward but make slow progress in
achieving sustainable development goals. Environmental systems experience degradation,
although there are some improvements and overall the intensity of resource and energy
use declines. Global population growth is moderate and levels off in the second half of
the century. Income inequality persists or improves only slowly and challenges to
reducing vulnerability to societal and environmental changes remain.""",
    ),
    dict(
        id="3",
        name="Regional Rivalry",
        name_long="Regional Rivalry – A Rocky Road",
        description="""High challenges to mitigation and adaptation.

A resurgent nationalism, concerns about competitiveness and security, and regional
conflicts push countries to increasingly focus on domestic or, at most, regional issues.
Policies shift over time to become increasingly oriented toward national and regional
security issues. Countries focus on achieving energy and food security goals within
their own regions at the expense of broader-based development. Investments in education
and technological development decline. Economic development is slow, consumption is
material-intensive, and inequalities persist or worsen over time. Population growth is
low in industrialized and high in developing countries. A low international priority for
addressing environmental concerns leads to strong environmental degradation in some
regions.""",
    ),
    dict(
        id="4",
        name="Inequality",
        name_long="Inequality – A Road Divided",
        description="""Low challenges to mitigation, high challenges to adaptation.

Highly unequal investments in human capital, combined with increasing disparities in
economic opportunity and political power, lead to increasing inequalities and
stratification both across and within countries. Over time, a gap widens between an
internationally-connected society that contributes to knowledge- and capital-intensive
sectors of the global economy, and a fragmented collection of lower-income, poorly
educated societies that work in a labor intensive, low-tech economy. Social cohesion
degrades and conflict and unrest become increasingly common. Technology development is
high in the high-tech economy and sectors. The globally connected energy sector
diversifies, with investments in both carbon-intensive fuels like coal and
unconventional oil, but also low-carbon energy sources. Environmental policies focus on
local issues around middle and high income areas.""",
    ),
    dict(
        id="5",
        name="Fossil-fueled Development",
        name_long="Fossil-fueled Development – Taking the Highway",
        description="""High challenges to mitigation, low challenges to adaptation.

This world places increasing faith in competitive markets, innovation and participatory
societies to produce rapid technological progress and development of human capital as
the path to sustainable development. Global markets are increasingly integrated. There
are also strong investments in health, education, and institutions to enhance human and
social capital. At the same time, the push for economic and social development is
coupled with the exploitation of abundant fossil fuel resources and the adoption of
resource and energy intensive lifestyles around the world. All these factors lead to
rapid growth of the global economy, while global population peaks and declines in the
21st century. Local environmental problems like air pollution are successfully managed.
There is faith in the ability to effectively manage social and ecological systems,
including by geo-engineering if necessary.""",
    ),
)


def generate(context: "Context", base_dir: Optional["PathLike"] = None):
    """Generate SDMX code lists containing the SSPs."""
    # Agency for ICONICS as the maintainer of other objects
    ICONICS = m.Agency(
        id="ICONICS",
        name="International Committee on New Integrated Climate Change Assessment "
        "Scenarios",
        contact=[m.Contact(uri=["https://depts.washington.edu/iconics/"])],
    )
    register_agency(ICONICS)

    for cl_info in CL_INFO:
        # Create the codelist: format the name and description
        cl: m.Codelist = m.Codelist(
            id="SSP",
            name=f"Shared Socioeconomic Pathways ({cl_info['name_extra']})",
            description="\n".join(
                wrap(DESC.strip(), width=len(DESC)) + ["", cl_info["desc_extra"]]
            ),
            version=cl_info["version"],
            maintainer=ICONICS,
        )

        # Add one Code for each SSP
        for info in CODE_INFO:
            # Construct the original ID
            original_id = f"SSP{info['id']}"

            # Format the name, description; add an annotation
            c = m.Code(
                id=info["id"],
                name=f"{original_id}: {info['name']}",
                description="\n".join(
                    [f"{original_id}: {info['name_long']}", "", info["description"]]
                ),
                annotations=[m.Annotation(id="original-id", text=original_id)],
            )

            # Append to the code list
            cl.append(c)

            # Construct a URN
            c.urn = sdmx.urn.make(c, maintainable_parent=cl)

        if context.dry_run:
            log.info(f"(dry run) Would write:\n{repr(cl)}")
            continue

        write(cl, base_dir)


SSP = SSP_2017 = make_enum("ICONICS:SSP(2017)")
SSP_2024 = make_enum("ICONICS:SSP(2024)")



import logging
import re
from collections import ChainMap
from copy import copy
from functools import lru_cache
from itertools import product
from typing import Dict, List, Mapping, MutableMapping, Tuple

import click
import pandas as pd
import pycountry
import xarray as xr
from iam_units import registry
from sdmx.model.v21 import Annotation, Code, Codelist

from message_ix_models.util import eval_anno, load_package_data, package_data_path
from message_ix_models.util.sdmx import as_codes

log = logging.getLogger(__name__)


@lru_cache()
def codelists(kind: str) -> List[str]:
    """Return a valid IDs for code lists of `kind`.

    Parameters
    ----------
    kind : str
        "node" or "year".
    """
    return sorted(path.stem for path in package_data_path(kind).glob("*.yaml"))


@lru_cache()
def get_codes(name: str) -> List[Code]:
    """Return codes for the dimension/set `name` in MESSAGE-GLOBIOM scenarios.

    The information is read from :file:`data/{name}.yaml`, e.g.
    :file:`data/technology.yaml`.

    When `name` includes "node", then child codes are automatically populated from the
    ISO 3166 database via :mod:`pycountry`. For instance:

    .. code-block:: yaml

       myregion:
         name: Custom region
         child: [AUT, SCG]

    …results in a region with child codes for Austria (a current country) and the
    formerly-existing country Serbia and Montenegro.

    Parameters
    ----------
    name : :class:`str`
        Any :file:`.yaml` file in the folder :file:`message_ix_models/data/`.

    Returns
    -------
    list of :class:`~sdmx.model.Code`
        Every Code has :attr:`id`, :attr:`name`, :attr:`description`, and
        :attr:`annotations` attributes. Calling :func:`str` on a code returns its
        :attr:`id`.
    """
    # Raw contents of the config file
    config = load_package_data(name)

    if "node" in name:
        # Automatically add information for countries within regions in the node
        # codelists. Use a ChainMap to combine a the `config` loaded from file and then
        # fall back to contents of the pycountry databases.
        config = ChainMap(
            config,
            # Create codes using the ISO database via pycountry
            {c.alpha_3: dict(id=c.alpha_3, name=c.name) for c in pycountry.countries},
            # Also include historic countries
            {
                c.alpha_3: dict(id=c.alpha_3, name=c.name)
                for c in pycountry.historic_countries
            },
        )

    # Convert to codes
    data = as_codes(config)

    # Fill in additional data, defaults, etc.
    if name == "commodity":
        process_commodity_codes(data)
    elif name == "technology":
        process_technology_codes(data)

    return data


@lru_cache()
def get_codelist(name: str) -> Codelist:
    """Return a |Codelist| for the dimension/set `name` in MESSAGE-GLOBIOM scenarios."""
    cl = Codelist(id=name.replace("/", "_").upper())
    cl.extend(get_codes(name))
    return cl


@lru_cache()
def get_region_codes(codelist: str) -> List[Code]:
    """Return the codes that are children of "World" in the specified `codelist`."""
    nodes = get_codes(f"node/{codelist}")
    return nodes[nodes.index(Code(id="World"))].child


def generate_product(
    data: Mapping, name: str, template: Code
) -> Tuple[List[Code], Dict[str, xr.DataArray]]:
    """Generates codes using a `template` by Cartesian product along ≥1 dimensions.

    :func:`generate_set_elements` is called for each of the `dims`, and these values
    are used to format `base`.

    Parameters
    ----------
    data
        Mapping from dimension IDs to lists of codes.
    name : str
        Name of the set.
    template : Code
        Must have Python format strings for its its :attr:`id` and :attr:`name`
        attributes.
    """
    # eval() and remove the original annotation
    dims = eval_anno(template, "_generate")
    template.pop_annotation(id="_generate")

    def _base(dim, match):
        """Return codes along dimension `dim`.

        If `match` is given, only children matching an expression."""
        dim_codes = data[dim]["add"]

        try:
            i = dim_codes.index(match)
        except ValueError:
            if isinstance(match, str):
                expr = re.compile(match)
                dim_codes = list(filter(lambda c: expr.match(c.id), dim_codes))
        else:
            dim_codes = dim_codes[i].child

        return dim_codes

    codes = []  # Accumulate codes and indices
    indices = []

    # Iterate over the product of filtered codes for each dimension in
    for item in product(*[_base(*dm) for dm in dims.items()]):
        result = copy(template)  # Duplicate the template

        fmt = dict(zip(dims.keys(), item))  # Format the ID and name
        result.id = result.id.format(**fmt)
        result.name = str(result.name).format(**fmt)  # type: ignore [assignment]

        codes.append(result)  # Store code and indices
        indices.append(tuple(map(str, item)))

    # - Convert length-N sequence of D-tuples to D iterables each of length N.
    # - Convert to D × 1-dimensional xr.DataArrays, each of length N.
    tmp = zip(*indices)
    indexers = {d: xr.DataArray(list(i), dims=name) for d, i in zip(dims.keys(), tmp)}
    # Corresponding indexer with the full code IDs
    indexers[name] = xr.DataArray([c.id for c in codes], dims=name)

    return codes, indexers


def generate_set_elements(data: MutableMapping, name) -> None:
    """Generate elements for set `name`.

    This function converts lists of codes in `data`, calling :func:`generate_product`
    and :func:`process_units_anno` as appropriate.

    Parameters
    ----------
    data
        Mapping from dimension IDs to lists of codes.
    name : str
        Name of the set for which to generate elements e.g. "commodity" or "technology".
    """
    hierarchical = name in {"technology"}

    codes = []  # Accumulate codes
    deferred = []
    for code in as_codes(data[name].get("add", [])):
        if name in {"commodity", "technology"}:
            process_units_anno(name, code, quiet=True)

        if eval_anno(code, "_generate"):
            # Requires a call to generate_product(); do these last
            deferred.append(code)
            continue

        codes.append(code)

        if hierarchical:
            # Store the children of `code`
            codes.extend(filter(lambda c: c not in codes, code.child))

    # Store codes processed so far, in case used recursively by generate_product()
    data[name]["add"] = codes

    # Use generate_product() to generate codes and indexers based on other sets
    for code in deferred:
        generated, indexers = generate_product(data, name, code)

        # Store
        data[name]["add"].extend(generated)

        # NB if there are >=2 generated groups, only indexers for the last are kept
        data[name]["indexers"] = indexers


def process_units_anno(set_name: str, code: Code, quiet: bool = False) -> None:
    """Process an annotation on `code` with id="units".

    The annotation text is wrapped as ``'registry.Unit("{text}")'``, such that it can
    be retrieved with :func:`.eval_anno` or :meth:`.ScenarioInfo.units_for`. If `code`
    has direct children, the annotation is also copied to those codes.

    Parameters
    ----------
    set_name : str
        Used in logged messages when `quiet` is :data:`False`.
    quiet : bool, *optional*
        If :data:`False` (the default), log on level :ref:`WARNING <python:levels>` if:

        - the annotation is missing, or
        - its text is not parseable with the :mod:`pint` application registry, i.e.
          :data:`iam_units.registry`.

        Otherwise, log on :ref:`DEBUG <python:levels>`.

    """
    level = logging.NOTSET if quiet else logging.WARNING
    # Convert a "units" annotation to a code snippet that will return a pint.Unit
    # via eval_anno()
    try:
        units_anno = code.get_annotation(id="units")
    except KeyError:
        log.log(level, f"{set_name.title()} {code} lacks defined units")
        return

    # First try the expression as-is, in case already processed
    expr = None
    for candidate in (str(units_anno.text), f'registry.Unit("{units_anno.text}")'):
        # Check that the unit can be parsed by the pint.UnitRegistry
        try:
            result = eval(candidate)
        except Exception:
            continue
        else:
            if isinstance(result, registry.Unit):
                expr = candidate
                break

    if not expr:  # pragma: no cover
        # No coverage: code that triggers this exception should never be committed
        log.log(
            level,
            f"Unit '{units_anno.text}' for {set_name} {code} not pint compatible",
        )
    else:
        # Modify the annotation so eval_anno() can be used
        units_anno.text = expr

    # Also annotate child codes
    for c in code.child:
        c.annotations.append(copy(units_anno))


def process_commodity_codes(codes):
    """Process a list of codes for ``commodity``.

    The function warns for commodities missing units or with non-:mod:`pint`-compatible
    units.
    """
    for code in codes:
        # FIXME remove quiet=True; instead improve commodity.yaml with units
        process_units_anno("commodity", code, quiet=True)


def process_technology_codes(codes):
    """Process a list of codes for ``technology``.

    This function ensures every code has an annotation with id "vintaged", default
    :obj:`False`.
    """
    for code in codes:
        # FIXME remove quiet=True; instead improve technology.yaml with units
        process_units_anno("technology", code, quiet=True)

        try:
            anno = code.pop_annotation(id="vintaged")
        except KeyError:
            # Default value for 'vintaged'
            anno = Annotation(id="vintaged", text=repr(False))

        code.annotations.append(anno)


@click.command(name="techs")
@click.pass_obj
def cli(ctx):
    """Export metadata to technology.csv.

    This command transforms the technology metadata from the YAML file to CSV format.
    """
    # Convert each code to a pd.Series
    data = []
    for code in get_codes("technology"):
        # Base attributes
        d = dict(id=code.id, name=str(code.name), description=str(code.description))

        # Annotations
        for anno in ("type", "vintaged", "sector", "input", "output"):
            try:
                d[anno] = str(code.get_annotation(id=anno).text)
            except KeyError:
                pass

        data.append(pd.Series(d))

    # Combine series to a data frame
    techs = pd.DataFrame(data)

    # Write to file
    dest = ctx.get_local_path("technology.csv")
    print(f"Write to {dest}")

    techs.to_csv(dest, index=None, header=True)

    # Print the first few items of the data frame
    print(techs.head())



import pytest

from message_ix_models.model import Config


class TestConfig:
    @pytest.mark.parametrize(
        "values",
        [
            dict(regions="R11", years="A"),
            pytest.param(
                dict(regions="R99", years="A"),
                marks=pytest.mark.xfail(
                    raises=ValueError, reason="regions='R99' not among […]"
                ),
            ),
            pytest.param(
                dict(regions="R11", years="C"),
                marks=pytest.mark.xfail(
                    raises=ValueError, reason="regions='C' not among ['A', 'B']"
                ),
            ),
        ],
    )
    def test_check(self, values):
        c = Config(**values)
        c.check()



import logging
import re
from typing import Union

import click

from message_ix_models.util.click import common_params

from .structure import SSP, SSP_2017, SSP_2024, generate

__all__ = [
    "SSP",
    "SSP_2017",
    "SSP_2024",
    "generate",
    "parse",
    "ssp_field",
]

log = logging.getLogger(__name__)


def parse(value: Union[str, SSP_2017, SSP_2024]) -> Union[SSP_2017, SSP_2024]:
    """Parse `value` to a member of :data:`SSP_2017` or :data:`SSP_2024`."""
    if isinstance(value, (SSP_2017, SSP_2024)):
        return value

    log.debug(f"Assume {value!r} is from {SSP_2017}")

    if isinstance(value, str):
        return SSP_2017[re.sub("SSP([12345])", r"\1", value)]
    else:
        return SSP_2017(value)


class ssp_field:
    """SSP field for use in data classes."""

    def __init__(self, default: Union[SSP_2017, SSP_2024]):
        self._default = default

    def __set_name__(self, owner, name):
        self._name = "_" + name

    def __get__(self, obj, type) -> Union[SSP_2017, SSP_2024]:
        if obj is None:
            return None  # type: ignore [return-value]

        try:
            return obj.__dict__[self._name]
        except KeyError:
            return obj.__dict__.setdefault(self._name, self._default)

    def __set__(self, obj, value):
        if value is None:
            value = self._default
        setattr(obj, self._name, parse(value))


@click.group("ssp")
def cli():
    pass


@cli.command("gen-structures")
@common_params("dry_run")
@click.pass_obj
def gen_structures(context, **kwargs):
    """(Re)Generate the SSP data structures in SDMX."""
    generate(context)



from functools import lru_cache
from typing import Optional

from pycountry import countries, historic_countries

#: Mapping from common, non-standard country names to ISO 3166-1 names.
COUNTRY_NAME = {
    "Korea": "Korea, Republic of",
    "Republic of Korea": "Korea, Republic of",
    "South Korea": "Korea, Republic of",
    "Russia": "Russian Federation",
}


@lru_cache(maxsize=2**9)
def iso_3166_alpha_3(name: str) -> Optional[str]:
    """Return an ISO 3166 alpha-3 code for a country `name`.

    Parameters
    ----------
    name : str
        Country name. This is looked up in the `pycountry
        <https://pypi.org/project/pycountry/#countries-iso-3166>`_ 'name',
        'official_name', or 'common_name' field. Values in :data:`COUNTRY_NAME` are
        supported.

    Returns
    -------
    str or None
    """
    # Maybe map a known, non-standard value to a standard value
    name = COUNTRY_NAME.get(name, name)

    # Use pycountry's built-in, case-insensitive lookup on all fields including name,
    # official_name, and common_name
    for db in (countries, historic_countries):
        try:
            return db.lookup(name).alpha_3
        except LookupError:
            continue  # Not found in `db`, e.g. countries; try again

    return None



"""Generic tools for working with exogenous data sources."""
import logging
from abc import ABC, abstractmethod
from operator import itemgetter
from pathlib import Path
from typing import Any, Dict, Literal, Mapping, Optional, Tuple, Type

from genno import Computer, Key, Quantity, quote

from message_ix_models import ScenarioInfo
from message_ix_models.model.structure import get_codes
from message_ix_models.util import cached

__all__ = [
    "MEASURES",
    "SOURCES",
    "DemoSource",
    "ExoDataSource",
    "iamc_like_data_for_query",
    "prepare_computer",
    "register_source",
]

log = logging.getLogger(__name__)

#: Supported measures. Subclasses of :class:`.ExoDataSource` may provide support for
#: other measures.
#:
#: .. todo:: Store this in a separate code list or concept scheme.
MEASURES = ("GDP", "POP")

#: Known sources for data. Use :func:`register_source` to add to this collection.
SOURCES: Dict[str, Type["ExoDataSource"]] = {}


class ExoDataSource(ABC):
    """Base class for sources of exogenous data."""

    #: Identifier for this particular source.
    id: str = ""

    #: Optional name for the returned :class:`.Key`/:class:`.Quantity`. If not set by
    #: :meth:`.__init__`, then the "measure" keyword is used.
    name: str = ""

    #: Optional additional dimensions for the returned :class:`.Key`/:class:`.Quantity`.
    #: If not set by :meth:`.__init__`, the dimensions are :math:`(n, y)`.
    extra_dims: Tuple[str, ...] = ()

    @abstractmethod
    def __init__(self, source: str, source_kw: Mapping) -> None:
        """Handle `source` and `source_kw`.

        An implementation **must**:

        - Raise :class:`ValueError` if it does not recognize or cannot handle the
          arguments in `source` or `source_kw`.
        - Recognize and handle (if possible) a "measure" keyword in `source_kw` from
          :data:`MEASURES`.

        It **may**:

        - Transform these into other values, for instance by mapping certain values to
          others, applying regular expressions, or other operations.
        - Store those values as instance attributes for use in :meth:`__call__`.
        - Set :attr:`name` and/or :attr:`extra_dims` to control the behaviour of
          :func:`.prepare_computer`.
        - Log messages that give information that may help to debug a
          :class:`ValueError` for `source` or `source_kw` that cannot be handled.

        It **should not** actually load data or perform any time- or memory-intensive
        operations; these should only be triggered by :meth:`.__call__`.
        """
        raise ValueError

    @abstractmethod
    def __call__(self) -> Quantity:
        r"""Return the data.

        The Quantity returned by this method **must** have dimensions
        :math:`(n, y) \cup \text{extra_dims}`. If the original/upstream/raw data has
        different dimensionality (fewer or more dimensions; different dimension IDs),
        the code **must** transform these, make appropriate selections, etc.
        """
        raise NotImplementedError

    def transform(self, c: "Computer", base_key: Key) -> Key:
        """Prepare `c` to transform raw data from `base_key`.

        `base_key` identifies the :class:`.Quantity` that is returned by
        :meth:`.__call__`. Before the data is returned, :meth:`.transform` allows the
        data source to add additional tasks or computations to `c` that further
        transform the data. (These operations **may** be done in :meth:`.__call__`
        directly, but :meth:`.transform` allows use of other :mod:`genno` operators and
        conveniences.)

        The default implementation:

        1. Aggregates the data (:func:`.genno.operator.aggregate`) on the |n| dimension
           using "n::groups".
        2. Interpolates the data (:func:`.genno.operator.interpolate`) on the |y|
           dimension using "y::coords".
        """
        # Aggregate
        c.add(base_key + "1", "aggregate", base_key, "n::groups", keep=False)

        # Interpolate to the desired set of periods
        kw = dict(fill_value="extrapolate")
        k2 = base_key + "2"
        c.add(k2, "interpolate", base_key + "1", "y::coords", kwargs=kw)

        return k2


def prepare_computer(
    context,
    c: "Computer",
    source="test",
    source_kw: Optional[Mapping] = None,
    *,
    strict: bool = True,
) -> Tuple[Key, ...]:
    """Prepare `c` to compute GDP, population, or other exogenous data.

    Check each :class:`ExoDataSource` in :data:`SOURCES` to determine whether it
    recognizes and can handle `source` and `source_kw`. If a source is identified, add
    tasks to `c` that retrieve and process data into a :class:`.Quantity` with, at
    least, dimensions :math:`(n, y)`.

    Parameters
    ----------
    source : str
        Identifier of the source, possibly with other information to be handled by a
        :class:`ExoDataSource`.
    source_kw : dict, *optional*
        Keyword arguments for a Source class. These can include indexers, selectors, or
        other information needed by the source class to identify the data to be
        returned.

        If the key "measure" is present, it **should** be one of :data:`MEASURES`.
    strict : bool, *optional*
        Raise an exception if any of the keys to be added already exist.

    Returns
    -------
    tuple of .Key

    Raises
    ------
    ValueError
        if no source is registered which can handle `source` and `source_kw`.
    """
    # Handle arguments
    source_kw = source_kw or dict()
    if measure := source_kw.get("measure"):
        if measure not in MEASURES:
            log.warning(f"source keyword {measure = } not in recognized {MEASURES}")
    else:
        measure = "UNKNOWN"

    # Look up input data flow
    source_obj = None
    for cls in SOURCES.values():
        try:
            # Instantiate a Source object to provide this data
            source_obj = cls(source, source_kw or dict())
        except Exception:
            pass  # Class does not recognize the arguments

    if source_obj is None:
        raise ValueError(f"No source found that can handle {source!r}")

    # Add structural information to the Computer
    c.require_compat("message_ix_models.report.operator")

    # Retrieve the node codelist
    c.add("n::codes", quote(get_codes(f"node/{context.model.regions}")), strict=strict)

    # Convert the codelist into a nested dict for aggregate()
    c.add("n::groups", "codelist_to_groups", "n::codes", strict=strict)

    # Add information about the list of periods
    if "y" not in c:
        info = ScenarioInfo()
        info.year_from_codes(get_codes(f"year/{context.model.years}"))

        c.add("y", quote(info.Y))

    if "y0" not in c:
        c.add("y0", itemgetter(0), "y")

    # Above as coords/indexers
    c.add("y::coords", lambda years: dict(y=years), "y")
    c.add("y0::coord", lambda year: dict(y=year), "y0")

    # Retrieve the raw data
    k = Key(source_obj.name or measure.lower(), ("n", "y") + source_obj.extra_dims)
    k_raw = k + source_obj.id  # Tagged with the source ID
    keys = [k]  # Keys to return

    c.add(k_raw, source_obj)

    # Allow the class to further transform the data. See ExoDataSource.transform() for
    # the default: aggregate, then interpolate
    key = source_obj.transform(c, k_raw)

    # Alias `key` -> `k`
    c.add(k, key)

    # Index to y0
    k_y0 = k + "y0_indexed"
    c.add(k_y0, "index_to", k, "y0::coord")
    keys.append(k_y0)

    # TODO also insert (1) index to a particular label on the "n" dimension (2) both

    return tuple(keys)


def register_source(cls: Type[ExoDataSource]) -> Type[ExoDataSource]:
    """Register :class:`.ExoDataSource` `cls` as a source of exogenous data."""
    if cls.id in SOURCES:
        raise ValueError(f"{SOURCES[cls.id]} already registered for id {cls.id!r}")
    SOURCES[cls.id] = cls
    return cls


@register_source
class DemoSource(ExoDataSource):
    """Example source of exogenous population and GDP data.

    Parameters
    ----------
    source : str
        **Must** be like ``test s1``, where "s1" is a scenario ID from ("s0"…"s4").
    source_kw : dict
        **Must** contain an element "measure", one of :data:`MEASURES`.
    """

    id = "DEMO"

    def __init__(self, source, source_kw):
        if not source.startswith("test "):
            # Don't recognize this `source` string → can't provide data
            raise ValueError

        # Select the data according to the `source`; in this case, scenario
        *parts, scenario = source.partition("test ")
        self.indexers = dict(s=scenario)

        # Map from the measure ID to a variable name
        self.indexers.update(
            v={"POP": "Population", "GDP": "GDP"}[source_kw["measure"]]
        )

    def __call__(self) -> Quantity:
        from genno.operator import select

        # - Retrieve the data.
        # - Apply the prepared indexers.
        return self.random_data().pipe(select, self.indexers, drop=True)

    @staticmethod
    def random_data():
        """Generate some random data with n, y, s, and v dimensions."""
        from genno.operator import relabel
        from genno.testing import random_qty
        from pycountry import countries

        return random_qty(dict(n=len(countries), y=2, s=5, v=2), units="kg").pipe(
            relabel,
            n={f"n{i}": c.alpha_3 for i, c in enumerate(countries)},
            v={"v0": "Population", "v1": "GDP"},
            y={"y0": 2010, "y1": 2050},
        )


@cached
def iamc_like_data_for_query(
    path: Path,
    query: str,
    *,
    archive_member: Optional[str] = None,
    non_iso_3166: Literal["keep", "discard"] = "discard",
    replace: Optional[dict] = None,
) -> Quantity:
    """Load data from `path` in IAMC-like format and transform to :class:`.Quantity`.

    The steps involved are:

    1. Read the data file; use pyarrow for better performance.
    2. Immediately apply `query` to reduce the data to be handled in subsequent steps.
    3. Assert that Model, Scenario, Variable, and Unit are unique; store the unique
       values. This means that `query` **must** result in data with unique values for
       these dimensions.
    4. Transform "Region" labels to ISO 3166-1 alpha-3 codes using
       :func:`.iso_3166_alpha_3`.
    5. Drop entire time series without such codes; for instance "World".
    6. Transform to a pd.Series with "n" and "y" index levels; ensure the latter are
       int.
    7. Transform to :class:`.Quantity` with units.

    The result is :obj:`.cached`.

    Parameters
    ----------
    archive_member : bool, optional
        If given, `path` may be an archive with 2 or more members. The member named by
        `archive_member` is extracted and read.
    non_iso_3166 : bool, optional
        If "discard" (default), "region" labels that are not ISO 3166-1 country names
        are discarded, along with associated data. If "keep", such labels are kept.
    """
    import pandas as pd

    from message_ix_models.util.pycountry import iso_3166_alpha_3

    unique = dict()

    def drop_unique(df, names) -> pd.DataFrame:
        if len(df) == 0:
            raise RuntimeError(f"0 rows matching {query!r}")

        names_list = names.split()
        for name in names_list:
            values = df[name].unique()
            if len(values) > 1:
                raise RuntimeError(f"Not unique {name!r}: {values}")
            unique[name] = values[0]
        return df.drop(names_list, axis=1)

    def assign_n(df: pd.DataFrame) -> pd.DataFrame:
        if non_iso_3166 == "discard":
            return df.assign(n=df["REGION"].apply(iso_3166_alpha_3))
        else:
            return df.assign(n=df["REGION"].apply(lambda v: iso_3166_alpha_3(v) or v))

    # Identify the source object/buffer to read from
    if archive_member:
        # A single member in a ZIP archive that has >1 members
        import zipfile

        zf = zipfile.ZipFile(path)
        source: Any = zf.open(archive_member)
    else:
        # A direct path, possibly compressed
        source = path

    tmp = (
        pd.read_csv(source, engine="pyarrow")
        .query(query)
        .replace(replace or {})
        .rename(columns=lambda c: c.upper())
        .pipe(drop_unique, "MODEL SCENARIO VARIABLE UNIT")
        .pipe(assign_n)
        .dropna(subset=["n"])
        .drop("REGION", axis=1)
        .set_index("n")
        .rename(columns=lambda y: int(y))
        .rename_axis(columns="y")
        .stack()
        .dropna()
    )
    return Quantity(tmp, units=unique["UNIT"])



"""Atomic reporting operations for MESSAGEix-GLOBIOM."""
import itertools
import logging
import re
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Set, Tuple, Union

import ixmp
import pandas as pd
from genno import Quantity
from genno.core.operator import Operator
from genno.operator import pow
from iam_units import convert_gwp
from iam_units.emissions import SPECIES

from message_ix_models import Context
from message_ix_models.util import add_par_data

if TYPE_CHECKING:
    from pathlib import Path

    from genno import Computer, Key
    from sdmx.model.v21 import Code

log = logging.getLogger(__name__)

__all__ = [
    "add_par_data",
    "codelist_to_groups",
    "compound_growth",
    "exogenous_data",
    "filter_ts",
    "from_url",
    "get_ts",
    "gwp_factors",
    "make_output_path",
    "model_periods",
    "remove_ts",
    "share_curtailment",
]


def codelist_to_groups(
    codes: List["Code"], dim: str = "n"
) -> Mapping[str, Mapping[str, List[str]]]:
    """Convert `codes` into a mapping from parent items to their children.

    The returned value is suitable for use with :func:`genno.operator.aggregate`.

    If this is a list of nodes per :func:`.get_codes`, then the mapping is from regions
    to the ISO 3166-1 alpha-3 codes of the countries within each region. The code for
    the region itself is also included in the values to be aggregated, so that already-
    aggregated data will pass through.
    """

    groups = dict()
    for code in filter(lambda c: len(c.child), codes):
        groups[code.id] = [code.id] + list(map(str, code.child))

    return {dim: groups}


def compound_growth(qty: Quantity, dim: str) -> Quantity:
    """Compute compound growth along `dim` of `qty`."""
    # Compute intervals along `dim`
    # The value at index d is the duration between d and the next index d+1
    c = qty.coords[dim]
    dur = (c - c.shift({dim: 1})).fillna(0).shift({dim: -1}).fillna(0)
    # - Raise the values of `qty` to the power of the duration.
    # - Compute cumulative product along `dim` from the first index.
    # - Shift, so the value at index d is the growth relative to the prior index d-1
    # - Fill in 1.0 for the first index.
    return pow(qty, Quantity(dur)).cumprod(dim).shift({dim: 1}).fillna(1.0)


@Operator.define()
def exogenous_data():
    """No action.

    This exists to connect :func:`.exo_data.prepare_computer` to
    :meth:`genno.Computer.add`.
    """
    pass  # pragma: no cover


@exogenous_data.helper
def add_exogenous_data(
    func, c: "Computer", *, context=None, source=None, source_kw=None
) -> Tuple["Key", ...]:
    """Prepare `c` to compute exogenous data from `source`."""
    from message_ix_models.tools.exo_data import prepare_computer

    return prepare_computer(
        context or Context.get_instance(-1), c, source=source, source_kw=source_kw
    )


def filter_ts(df: pd.DataFrame, expr: re.Pattern, *, column="variable") -> pd.DataFrame:
    """Filter time series data in `df`.

    1. Keep only rows in `df` where `expr` is a full match (
       :meth:`~pandas.Series.str.fullmatch`) for the entry in `column`.
    2. Retain only the first match group ("...(...)...") from `expr` as the `column`
       entry.
    """
    return df[df[column].str.fullmatch(expr)].assign(
        variable=df[column].str.replace(expr, r"\1", regex=True)
    )


def get_ts(
    scenario: ixmp.Scenario,
    filters: Optional[dict] = None,
    iamc: bool = False,
    subannual: Union[bool, str] = "auto",
):
    """Retrieve timeseries data from `scenario`.

    Corresponds to :meth:`ixmp.Scenario.timeseries`.

    .. todo:: Move upstream, e.g. to :mod:`ixmp` alongside :func:`.store_ts`.
    """
    filters = filters or dict()

    return scenario.timeseries(iamc=iamc, subannual=subannual, **filters)


def gwp_factors() -> Quantity:
    """Use :mod:`iam_units` to generate a Quantity of GWP factors.

    The quantity is dimensionless, e.g. for converting [mass] to [mass], andhas
    dimensions:

    - 'gwp metric': the name of a GWP metric, e.g. 'SAR', 'AR4', 'AR5'. All metrics are
       on a 100-year basis.
    - 'e': emissions species, as in MESSAGE. The entry 'HFC' is added as an alias for
      the species 'HFC134a' from iam_units.
    - 'e equivalent': GWP-equivalent species, always 'CO2'.
    """
    dims = ["gwp metric", "e", "e equivalent"]
    metric = ["SARGWP100", "AR4GWP100", "AR5GWP100"]
    species_to = ["CO2"]  # Add to this list to perform additional conversions

    data = []
    for m, s_from, s_to in itertools.product(metric, SPECIES, species_to):
        # Get the conversion factor from iam_units
        factor = convert_gwp(m, (1, "kg"), s_from, s_to).magnitude

        # MESSAGEix-GLOBIOM uses e='HFC' to refer to this species
        if s_from == "HFC134a":
            s_from = "HFC"

        # Store entry
        data.append((m[:3], s_from, s_to, factor))

    # Convert to Quantity object and return
    return Quantity(
        pd.DataFrame(data, columns=dims + ["value"]).set_index(dims)["value"].dropna()
    )


def make_output_path(config: Mapping, name: str) -> "Path":
    """Return a path under the "output_dir" Path from the reporter configuration."""
    return config["output_dir"].joinpath(name)


def model_periods(y: List[int], cat_year: pd.DataFrame) -> List[int]:
    """Return the elements of `y` beyond the firstmodelyear of `cat_year`.

    .. todo:: Move upstream, to :mod:`message_ix`.
    """
    y0 = cat_year.query("type_year == 'firstmodelyear'")["year"].item()
    return list(filter(lambda year: y0 <= year, y))


def remove_ts(
    scenario: ixmp.Scenario,
    config: Optional[dict] = None,
    after: Optional[int] = None,
    dump: bool = False,
) -> None:
    """Remove all time series data from `scenario`.

    Note that data stored with :meth:`.add_timeseries` using :py:`meta=True` as a
    keyword argument cannot be removed using :meth:`.TimeSeries.remove_timeseries`, and
    thus also not with this operator.

    .. todo:: Move upstream, to :mod:`ixmp` alongside :func:`.store_ts`.
    """
    if dump:
        raise NotImplementedError

    data = scenario.timeseries().drop("value", axis=1)
    N = len(data)
    count = f"{N}"

    if after:
        query = f"{after} <= year"
        data = data.query(query)
        count = f"{len(data)} of {N} ({query})"

    log.info(f"Remove {count} rows of time series data from {scenario.url}")

    # TODO improve scenario.transact() to allow timeseries_only=True; use here
    scenario.check_out(timeseries_only=True)
    try:
        scenario.remove_timeseries(data)
    except Exception:
        scenario.discard_changes()
    else:
        scenario.commit(f"Remove time series data ({__name__}.remove_all_ts)")


# Non-weak references to objects to keep them alive
_FROM_URL_REF: Set[Any] = set()


def from_url(url: str, cls=ixmp.TimeSeries) -> ixmp.TimeSeries:
    """Return a :class:`ixmp.TimeSeries` or subclass instance, given its `url`.

    .. todo:: Move upstream, to :mod:`ixmp.reporting`.

    Parameters
    ----------
    cls : type, *optional*
        Subclass to instantiate and return; for instance, |Scenario|.
    """
    ts, mp = cls.from_url(url)
    assert ts is not None
    _FROM_URL_REF.add(ts)
    _FROM_URL_REF.add(mp)
    return ts


# commented: currently unused
# def share_cogeneration(fraction, *parts):
#     """Deducts a *fraction* from the first of *parts*."""
#     return parts[0] - (fraction * sum(parts[1:]))


def share_curtailment(curt, *parts):
    """Apply a share of *curt* to the first of *parts*.

    If this is being used, it usually will indicate the need to split *curt* into
    multiple technologies; one for each of *parts*.
    """
    return parts[0] - curt * (parts[0] / sum(parts))



from typing import Mapping, Optional, Tuple
from warnings import catch_warnings, filterwarnings

import pandas as pd
from iam_units import registry


def series_of_pint_quantity(*args, **kwargs) -> pd.Series:
    """Suppress a spurious warning.

    Creating a :class:`pandas.Series` with a list of :class:`pint.Quantity` triggers a
    warning “The unit of the quantity is stripped when downcasting to ndarray,” even
    though the entire object is being stored and the unit is **not** stripped. This
    function suppresses this warning.
    """
    with catch_warnings():
        filterwarnings(
            "ignore",
            message="The unit of the quantity is stripped when downcasting to ndarray",
            module="pandas.core.dtypes.cast",
        )

        return pd.Series(*args, **kwargs)


def convert_units(
    s: pd.Series,
    unit_info: Mapping[str, Tuple[float, str, Optional[str]]],
    store="magnitude",
) -> pd.Series:
    """Convert units of `s`, for use with :meth:`~pandas.DataFrame.apply`.

    ``s.name`` is used to retrieve a tuple of (`factor`, `input_unit`, `output_unit`)
    from `unit_info`. The (:class:`float`) values of `s` are converted to
    :class:`pint.Quantity` with the `input_unit` and factor; then cast to `output_unit`,
    if provided.

    Parameters
    ----------
    s : pandas.Series
    unit_info : dict (str -> tuple)
        Mapping from quantity name (matched to ``s.name``) to 3-tuples of (`factor`,
        `input_unit`, `output_unit`). `output_unit` may be :obj:`None`. For example,
        see :data:`.ikarus.UNITS`.
    store : "magnitude" or "quantity"
        If "magnitude", the values of the returned series are the magnitudes of the
        results, with no output units. If "quantity", the values are scalar
        :class:`~pint.Quantity` objects.

    Returns
    -------
    pandas.Series
        Same shape, index, and values as `s`, with output units.
    """
    if store not in "magnitude quantity":
        raise ValueError(f"{store = }")

    # Retrieve the information from `unit_info`
    factor, unit_in, unit_out = unit_info[s.name]

    # Default: `unit_out` is the same as `unit_in`
    unit_out = unit_out or unit_in

    # - Convert the values to a pint.Quantity(array) with the input units
    # - Convert to output units
    # - According to `store`, either extract just the magnitude, or store scalar
    #   pint.Quantity objects.
    # - Reassemble into a series with index matching `s`
    result = registry.Quantity(factor * s.values, unit_in).to(unit_out)

    return series_of_pint_quantity(
        result.magnitude if store == "magnitude" else result.tolist(),
        index=s.index,
        dtype=(float if store == "magnitude" else object),
        name=s.name,
    )



import logging
from functools import lru_cache, partial
from typing import Mapping

import pandas as pd
from sdmx.model.v21 import Code

from message_ix_models import ScenarioInfo
from message_ix_models.model import build
from message_ix_models.model.structure import get_codes
from message_ix_models.util import package_data_path

from .utils import read_config

log = logging.getLogger(__name__)


def get_spec(context) -> Mapping[str, ScenarioInfo]:
    """Return the specification for nexus implementation

    Parameters
    ----------
    context : .Context
        The key ``regions`` determines the regional aggregation used.
    """

    context = read_config()

    require = ScenarioInfo()
    remove = ScenarioInfo()
    add = ScenarioInfo()

    if context.nexus_set == "nexus":
        # Merge technology.yaml with set.yaml
        context["water set"]["nexus"]["technology"]["add"] = context[
            "water technology"
        ]["nexus"]
        # Update the ScenarioInfo objects with required and new set elements
        for set_name, config in context["water set"]["nexus"].items():
            # Required elements
            require.set[set_name].extend(config.get("require", []))

            # Elements to remove
            remove.set[set_name].extend(config.get("remove", []))

            # Elements to add
            add.set[set_name].extend(config.get("add", []))

        # The set of required nodes varies according to context.regions
        n_codes = get_codes(f"node/{context.regions}")
        nodes = list(map(str, n_codes[n_codes.index(Code(id="World"))].child))
        require.set["node"].extend(nodes)

        # Share commodity for groundwater
        results = {}
        df_node = context.all_nodes
        n = len(df_node.values)

        d = {
            "shares": ["share_low_lim_GWat"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_low_lim_GWat_share"] * n,
            "mode": ["M1"] * n,
            "commodity": ["groundwater_basin"] * n,
            "level": ["water_avail_basin"] * n,
        }

        df_share = pd.DataFrame(data=d)
        df_list = df_share.values.tolist()
        results["map_shares_commodity_share"] = df_list

        d = {
            "shares": ["share_low_lim_GWat"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_low_lim_GWat_total"] * n,
            "mode": ["M1"] * n,
            "commodity": ["surfacewater_basin"] * n,
            "level": ["water_avail_basin"] * n,
        }

        df_share = pd.DataFrame(data=d)

        d2 = {
            "shares": ["share_low_lim_GWat"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_low_lim_GWat_total"] * n,
            "mode": ["M1"] * n,
            "commodity": ["groundwater_basin"] * n,
            "level": ["water_avail_basin"] * n,
        }

        df_share2 = pd.DataFrame(data=d2)

        df_share = df_share.append(df_share2)
        df_list = df_share.values.tolist()

        results["map_shares_commodity_total"] = df_list

        for set_name, config in results.items():
            # Sets  to add
            add.set[set_name].extend(config)

        results = {}

        # Share commodity for urban water recycling
        d = {
            "shares": ["share_wat_recycle"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_wat_recycle_share"] * n,
            "mode": ["M1"] * n,
            "commodity": ["urban_collected_wst"] * n,
            "level": ["water_treat"] * n,
        }

        df_share = pd.DataFrame(data=d)
        df_list = df_share.values.tolist()
        results["map_shares_commodity_share"] = df_list

        d = {
            "shares": ["share_wat_recycle"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_wat_recycle_total"] * n,
            "mode": ["M1"] * n,
            "commodity": ["urban_collected_wst"] * n,
            "level": ["water_treat"] * n,
        }

        df_share = pd.DataFrame(data=d)

        d2 = {
            "shares": ["share_wat_recycle"] * n,
            "node_share": df_node,
            "node": df_node,
            "type_tec": ["share_wat_recycle_total"] * n,
            "mode": ["M1"] * n,
            "commodity": ["urban_collected_wst"] * n,
            "level": ["water_treat"] * n,
        }

        df_share2 = pd.DataFrame(data=d2)

        df_share = df_share.append(df_share2)
        df_list = df_share.values.tolist()

        results["map_shares_commodity_total"] = df_list

        for set_name, config in results.items():
            # Sets  to add
            add.set[set_name].extend(config)

    elif context.nexus_set == "cooling":
        # Merge technology.yaml with set.yaml
        context["water set"]["cooling"]["technology"]["add"] = context[
            "water technology"
        ]["cooling"]
        # Update the ScenarioInfo objects with required and new set elements
        for set_name, config in context["water set"]["cooling"].items():
            # Required elements
            require.set[set_name].extend(config.get("require", []))

            # Elements to remove
            remove.set[set_name].extend(config.get("remove", []))

            # Elements to add
            add.set[set_name].extend(config.get("add", []))

    return dict(require=require, remove=remove, add=add)

    return dict(require=require, remove=remove, add=add)


@lru_cache()
def generate_set_elements(set_name, match=None):
    codes = read_config()["water set"][set_name].get("add", [])

    hierarchical = set_name in {"technology"}

    results = []
    for code in codes:
        if match and code.id != match:
            continue
        elif hierarchical:
            results.extend(code)

    return results


def map_basin(context) -> Mapping[str, ScenarioInfo]:
    """Return specification for mapping basins to regions

    The basins are spatially consolidated from HydroSHEDS basins delineation
    database.This delineation is then intersected with MESSAGE regions to form new
    water sector regions for the nexus module.
    The nomenclature for basin names is <basin_id>|<MESSAGEregion> such as R1|AFR
    """
    context = read_config()

    add = ScenarioInfo()

    require = ScenarioInfo()

    remove = ScenarioInfo()

    # define an empty dictionary
    results = {}
    # read csv file for basin names and region mapping
    # reading basin_delineation
    FILE = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE)

    df = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df["node"] = "B" + df["BCU_name"].astype(str)
    df["mode"] = "M" + df["BCU_name"].astype(str)
    if context.type_reg == "country":
        df["region"] = context.map_ISO_c[context.regions]
    else:
        df["region"] = f"{context.regions}_" + df["REGION"].astype(str)

    results["node"] = df["node"]
    results["mode"] = df["mode"]
    # map nodes as per dimensions
    df1 = pd.DataFrame({"node_parent": df["region"], "node": df["node"]})
    df2 = pd.DataFrame({"node_parent": df["node"], "node": df["node"]})
    frame = [df1, df2]
    df_node = pd.concat(frame)
    nodes = df_node.values.tolist()

    results["map_node"] = nodes

    context.all_nodes = df["node"]

    for set_name, config in results.items():
        # Sets  to add
        add.set[set_name].extend(config)

    return dict(require=require, remove=remove, add=add)


def main(context, scenario, **options):
    """Set up MESSAGEix-Nexus on `scenario`.

    See also
    --------
    add_data
    apply_spec
    get_spec
    """
    from .data import add_data

    log.info("Set up MESSAGEix-Nexus")

    if context.nexus_set == "nexus":
        # Add water balance
        spec = map_basin(context)

        # Apply the structural changes AND add the data
        build.apply_spec(scenario, spec, **options)

    # Core water structure
    spec1 = get_spec(context)

    # Apply the structural changes AND add the data
    build.apply_spec(scenario, spec1, partial(add_data, context=context), **options)

    # Uncomment to dump for debugging
    # scenario.to_excel('debug.xlsx')



import pytest
from genno import Computer

from message_ix_models.project.ssp import (
    SSP,
    SSP_2017,
    SSP_2024,
    generate,
    parse,
    ssp_field,
)
from message_ix_models.project.ssp.data import SSPUpdate  # noqa: F401
from message_ix_models.tools.exo_data import prepare_computer


def test_generate(tmp_path, test_context):
    # Function runs
    generate(test_context, base_dir=tmp_path)

    # Two XML files are created
    assert 2 == len(list(tmp_path.glob("*.xml")))


def test_enum():
    # Enumerations have the expected length
    assert 5 == len(SSP_2017)
    assert 5 == len(SSP_2024)

    # Members can be accessed by ID
    a = SSP_2017["1"]

    # …or by value
    b = SSP_2017(1)

    # …all retrieving the same member
    assert a == b

    # __getattr__ lookup does not invoke _missing_
    with pytest.raises(KeyError):
        SSP_2017[1]

    # Same SSP ID from different enums are not equivalent
    assert SSP_2017["1"] != SSP_2024["1"]
    assert SSP_2017["1"] is not SSP_2024["1"]
    assert SSP["1"] != SSP_2024["1"]


@pytest.mark.parametrize(
    "expected, value",
    (
        (SSP_2017["1"], SSP_2017["1"]),  # Literal value
        (SSP_2017["1"], "1"),  # String ID as appears in the codelist
        (SSP_2017["1"], "SSP1"),  # Prefixed by "SSP"
        (SSP_2017["1"], 1),  # Integer
    ),
)
def test_parse(value, expected):
    assert expected == parse(value)


def test_ssp_field() -> None:
    from dataclasses import dataclass

    @dataclass
    class Foo:
        bar: ssp_field = ssp_field(default=SSP_2017["1"])
        baz: ssp_field = ssp_field(default=SSP_2024["5"])

    # Can be instantiated with no arguments
    f = Foo()
    assert SSP_2017["1"] is f.bar
    assert SSP_2024["5"] is f.baz

    # Can be instantiated with different values
    f = Foo(bar=SSP_2017["3"], baz=SSP_2024["3"])
    assert SSP_2017["3"] is f.bar
    assert SSP_2024["3"] is f.baz

    # Values can be set and are passed through parse()
    f.bar = "SSP2"
    f.baz = "SSP4"

    assert SSP_2017["2"] is f.bar
    assert SSP_2017["4"] is f.baz


def test_cli(mix_models_cli):
    mix_models_cli.assert_exit_0(["ssp", "gen-structures", "--dry-run"])


class TestSSPOriginal:
    @pytest.mark.parametrize(
        "source",
        (
            "ICONICS:SSP(2017).1",
            "ICONICS:SSP(2017).2",
            "ICONICS:SSP(2017).3",
            "ICONICS:SSP(2017).4",
            "ICONICS:SSP(2017).5",
        ),
    )
    @pytest.mark.parametrize(
        "source_kw",
        (
            dict(measure="POP", model="OECD Env-Growth"),
            dict(measure="GDP", model="OECD Env-Growth"),
            # Excess keyword arguments
            pytest.param(
                dict(measure="GDP", model="OECD Env-Growth", foo="bar"),
                marks=pytest.mark.xfail(raises=ValueError),
            ),
        ),
    )
    def test_prepare_computer(self, test_context, source, source_kw):
        # FIXME The following should be redundant, but appears mutable on GHA linux and
        #       Windows runners.
        test_context.model.regions = "R14"

        c = Computer()

        keys = prepare_computer(test_context, c, source, source_kw)

        # Preparation of data runs successfully
        result = c.get(keys[0])

        # Data has the expected dimensions
        assert ("n", "y") == result.dims

        # Data is complete
        assert 14 == len(result.coords["n"])
        assert 14 == len(result.coords["y"])


class TestSSPUpdate:
    @pytest.mark.parametrize(
        "source",
        (
            "ICONICS:SSP(2024).1",
            "ICONICS:SSP(2024).2",
            "ICONICS:SSP(2024).3",
            "ICONICS:SSP(2024).4",
            "ICONICS:SSP(2024).5",
        ),
    )
    @pytest.mark.parametrize(
        "source_kw",
        (
            dict(measure="POP"),
            dict(measure="GDP", model="IIASA GDP 2023"),
            dict(measure="GDP", model="OECD ENV-Growth 2023"),
            # Excess keyword arguments
            pytest.param(
                dict(measure="POP", foo="bar"),
                marks=pytest.mark.xfail(raises=ValueError),
            ),
        ),
    )
    def test_prepare_computer(self, test_context, source, source_kw):
        # FIXME The following should be redundant, but appears mutable on GHA linux and
        #       Windows runners.
        test_context.model.regions = "R14"

        c = Computer()

        keys = prepare_computer(test_context, c, source, source_kw)

        # Preparation of data runs successfully
        result = c.get(keys[0])

        # Data has the expected dimensions
        assert ("n", "y") == result.dims

        # Data is complete
        assert 14 == len(result.coords["n"])
        assert 14 == len(result.coords["y"])
